# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.data_inputs.data_input import DataInputAbstract
from montepy.input_parser.tally_parser import TallyParser


class TallyMultiplier(DataInputAbstract):
    _parser = TallyParser()

    @staticmethod
    def _class_prefix():
        return "fm"

    @staticmethod
    def _has_number():
        return True

    @staticmethod
    def _has_classifier():
        return 0



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy import mcnp_problem
from montepy.constants import DEFAULT_VERSION


def read_input(input_file, mcnp_version=DEFAULT_VERSION):
    """
    Reads the specified MCNP Input file.

    The MCNP version must be a three component tuple e.g., (6, 2, 0) and (5, 1, 60).

    :param input_file: the path to the input file to read.
    :type input_file: str
    :param mcnp_version: The version of MCNP that the input is intended for.
    :type mcnp_version: tuple
    :returns: The MCNP_Problem instance representing this file.
    :rtype: MCNP_Problem
    :raises UnsupportedFeature: If an input format is used that MontePy does not support.
    :raises MalformedInputError: If an input has a broken syntax.
    :raises NumberConflictError: If two objects use the same number in the input file.
    :raises BrokenObjectLinkError: If a reference is made to an object that is not in the input file.
    :raises UnknownElement: If an isotope is specified for an unknown element.
    """
    problem = mcnp_problem.MCNP_Problem(input_file)
    problem.mcnp_version = mcnp_version
    problem.parse_input()
    return problem



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from abc import abstractmethod
import montepy
from montepy.data_inputs.data_input import DataInputAbstract
from montepy.input_parser import syntax_node
from montepy.input_parser.block_type import BlockType
from montepy.input_parser.mcnp_input import Input, Jump
import warnings


class CellModifierInput(DataInputAbstract):
    """
    Abstract Parent class for Data Inputs that modify cells / geometry.

    Examples: IMP, VOL, etc.

    :param input: the Input object representing this data input
    :type input: Input
    :param in_cell_block: if this card came from the cell block of an input file.
    :type in_cell_block: bool
    :param key: the key from the key-value pair in a cell
    :type key: str
    :param value: the value syntax tree from the key-value pair in a cell
    :type value: SyntaxNode
    """

    def __init__(self, input=None, in_cell_block=False, key=None, value=None):
        fast_parse = False
        if key and value:
            input = Input([key], BlockType.DATA)
            fast_parse = True
        super().__init__(input, fast_parse)
        if not isinstance(in_cell_block, bool):
            raise TypeError("in_cell_block must be a bool")
        if key and not isinstance(key, str):
            raise TypeError("key must be a str")
        if value and not isinstance(value, syntax_node.SyntaxNode):
            raise TypeError("value must be from a SyntaxNode")
        if not in_cell_block and not input:
            self._generate_default_data_tree()
        self._in_cell_block = in_cell_block
        self._in_key = key
        self._in_value = value
        if key and in_cell_block:
            self._set_in_cell_block = True
            self._tree = value
            self._data = value["data"]
        else:
            self._set_in_cell_block = False
            if in_cell_block and key is None and value is None:
                self._generate_default_cell_tree()

    @abstractmethod
    def _generate_default_cell_tree(self):
        pass

    def _generate_default_data_tree(self):
        list_node = syntax_node.ListNode("number sequence")
        list_node.append(self._generate_default_node(float, None))
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = self._generate_default_node(
            str, self._class_prefix().upper(), None
        )
        classifier.padding = syntax_node.PaddingNode(" ")
        self._tree = syntax_node.SyntaxNode(
            self._class_prefix(),
            {
                "start_pad": syntax_node.PaddingNode(),
                "classifier": classifier,
                "keyword": syntax_node.ValueNode(None, str, None),
                "data": list_node,
            },
        )

    @property
    def in_cell_block(self):
        """
        True if this object represents an input from the cell block section of a file.

        :rtype: bool
        """
        return self._in_cell_block

    @property
    def set_in_cell_block(self):
        """
        True if this data were set in the cell block in the input
        """
        return self._set_in_cell_block

    @abstractmethod
    def merge(self, other):
        """
        Merges the data from another card of same type into this one.

        :param other: The other object to merge into this object.
        :type other: CellModifierInput
        :raises MalformedInputError: if two objects cannot be merged.
        """
        pass

    def link_to_problem(self, problem):
        super().link_to_problem(problem)
        if self.set_in_cell_block:
            self._problem.print_in_data_block[self._class_prefix()] = False

    @abstractmethod
    def push_to_cells(self):
        """
        After being linked to the problem update all cells attributes with this data.

        This needs to also check that none of the cells had data provided in the cell block
        (check that ``set_in_cell_block`` isn't set).
        Use ``self._check_redundant_definitions`` to do this.

        :raises MalformedInputError: When data are given in the cell block and the data block.
        """
        pass

    @property
    @abstractmethod
    def has_information(self):
        """
        For a cell instance of :class:`montepy.data_cards.cell_modifier.CellModifierCard` returns True iff there is information here worth printing out.

        e.g., a manually set volume for a cell

        :returns: True if this instance has information worth printing.
        :rtype: bool
        """
        pass

    def _check_redundant_definitions(self):
        """
        Checks that data wasn't given in data block and the cell block.
        """
        attr, _ = montepy.Cell._INPUTS_TO_PROPERTY[type(self)]
        if not self._in_cell_block and self._problem:
            cells = self._problem.cells
            for cell in cells:
                if getattr(cell, attr).set_in_cell_block:
                    raise montepy.errors.MalformedInputError(
                        cell._input,
                        f"Cell: {cell.number} provided {self._class_prefix().upper()}"
                        " data when those data were in the data block",
                    )

    @abstractmethod
    def _clear_data(self):
        """
        After data has been pushed to cells, delete internal data to avoid inadvertent editing.

        This is only called on data-block instances of this object.
        """
        pass

    @property
    def _is_worth_printing(self):
        """
        Determines if this object has information that is worth printing in the input file.

        Uses the :func:`has_information` property for all applicable cell(s)

        :returns: True if this object should be included in the output
        :rtype: bool
        """
        if self.in_cell_block:
            return self.has_information
        attr, _ = montepy.Cell._INPUTS_TO_PROPERTY[type(self)]
        for cell in self._problem.cells:
            if getattr(cell, attr).has_information:
                return True
        return False

    @property
    @abstractmethod
    def _tree_value(self):
        """
        The ValueNode that holds the information for this instance, that should be included in the data block.

        .. versionadded:: 0.2.0

        :returns: The ValueNode to update the data-block syntax tree with.
        :rtype: ValueNode
        """
        pass

    def _collect_new_values(self):
        """
        Gets a list of the ValueNodes that hold the information for all cells.

        This will be a list in the same order as :func:`montepy.mcnp_problem.MCNP_Problem.cells`.

        .. versionadded:: 0.2.0

        :returns: a list of the ValueNodes to update the data block syntax tree with
        :rtype: list
        """
        ret = []
        attr, _ = montepy.Cell._INPUTS_TO_PROPERTY[type(self)]
        for cell in self._problem.cells:
            input = getattr(cell, attr)
            ret.append(input._tree_value)
        return ret

    @abstractmethod
    def _update_cell_values(self):
        """
        Updates values in the syntax tree when in the cell block.

        .. versionadded:: 0.2.0
        """
        pass

    def _update_values(self):
        if self.in_cell_block:
            self._update_cell_values()
        else:
            new_vals = self._collect_new_values()
            self.data.update_with_new_values(new_vals)

    def _format_tree(self):
        """
        Formats the syntax tree for printing in an input file.

        By default this runs ``self._tree.format()``.

        :returns: a string of the text to write out to the input file (not wrapped yet for MCNP).
        :rtype: str
        """
        return self._tree.format()

    def format_for_mcnp_input(self, mcnp_version):
        """
        Creates a string representation of this MCNP_Object that can be
        written to file.

        :param mcnp_version: The tuple for the MCNP version that must be exported to.
        :type mcnp_version: tuple
        :return: a list of strings for the lines that this input will occupy.
        :rtype: list
        """
        self.validate()
        if not self._problem:
            print_in_data_block = not self.in_cell_block
        else:
            print_in_data_block = self._problem.print_in_data_block[
                self._class_prefix().upper()
            ]

        """
        Boolean logic
        A= self.in_Cell_block
        B = print_in_data_block
        C = is_worth_pring

        Logic:
            A!BC + !ABC
            C *(A!B + !AB)
            C * (A xor B)
            C * (A != B)
        """
        # print in either block
        if (self.in_cell_block != print_in_data_block) and self._is_worth_printing:
            self._update_values()
            return self.wrap_string_for_mcnp(self._format_tree(), mcnp_version, True)
        return []

    @property
    def has_changed_print_style(self):  # pragma: no cover
        """
        returns true if the printing style for this modifier has changed
        from cell block to data block, or vice versa.

        .. deprecated:: 0.2.0
            This property is no longer needed and overly complex.

        :returns: true if the printing style for this modifier has changed
        :rtype: bool
        :raises DeprecationWarning: raised always.
        """
        warnings.warn(
            "has_changed_print_style will be removed soon.",
            DeprecationWarning,
            stacklevel=2,
        )
        if self._problem:
            print_in_cell_block = not self._problem.print_in_data_block[
                self.class_prefix
            ]
            set_in_cell_block = print_in_cell_block
            if not self.in_cell_block:
                for cell in self._problem.cells:
                    attr = montepy.Cell._CARDS_TO_PROPERTY[type(self)][0]
                    modifier = getattr(cell, attr)
                    if modifier.has_information:
                        set_in_cell_block = modifier.set_in_cell_block
                    break
            else:
                if self.has_information:
                    set_in_cell_block = self.set_in_cell_block
            return print_in_cell_block ^ set_in_cell_block
        else:
            return False



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from enum import Enum, unique


@unique
class Particle(Enum):
    """
    Supported MCNP supported particles.

    Taken from Table 2-2 of LA-UR-17-29981.
    """

    NEUTRON = "N"
    PHOTON = "P"
    ELECTRON = "E"
    NEGATIVE_MUON = "|"
    ANTI_NEUTRON = "Q"
    ELECTRON_NEUTRINO = "U"
    MUON_NEUTRINO = "V"
    POSITRON = "F"
    PROTON = "H"
    LAMBDA_BARYON = "L"
    POSITIVE_SIGMA_BARYON = "+"
    NEGATIVE_SIGMA_BARYON = "-"
    CASCADE = "X"
    NEGATIVE_CASCADE = "Y"
    OMEGA_BARYON = "O"
    POSITIVE_MUON = "!"
    ANTI_ELECTRON_NEUTRINO = "<"
    ANTI_MUON_NEUTRINO = ">"
    ANTI_PROTON = "G"
    POSITIVE_PION = "/"
    NEUTRAL_PION = "Z"
    POSITIVE_KAON = "K"
    KAON_SHORT = "%"
    KAON_LONG = "^"
    ANTI_LAMBDA_BARYON = "B"
    ANTI_POSITIVE_SIGMA_BARYON = "_"
    ANTI_NEGATIVE_SIGMA_BARYON = "~"
    ANTI_CASCADE = "C"
    POSITIVE_CASCADE = "W"
    ANTI_OMEGA = "@"
    DEUTERON = "D"
    TRITON = "T"
    HELION = "S"
    ALPHA_PARTICLE = "A"
    NEGATIVE_PION = "*"
    NEGATIVE_KAON = "?"
    HEAVY_ION = "#"

    def __lt__(self, other):
        return self.value < other.value

    def __str__(self):
        return self.name.lower()



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.errors import UnsupportedFeature

"""
Constants related to how MCNP inputs are formatted, and MontePy behavior.
"""

rel_tol = 1e-9
"""
Relative tolerance passed to math.isclose.

:rtype: float
"""

abs_tol = 0.0
"""
Absolute tolerance passed to math.isclose.

:rtype: float
"""


BLANK_SPACE_CONTINUE = 5
"""
Number of spaces in a new line before it's considered a continuation.
"""

LINE_LENGTH = {(5, 1, 60): 80, (6, 1, 0): 80, (6, 2, 0): 128}
"""
The number of characters allowed in a line for each MCNP version.
"""

DEFAULT_VERSION = (6, 2, 0)
"""
The default version of MCNP to use.
"""

TABSIZE = 8
"""
How many spaces a tab is expand to.
"""


def get_max_line_length(mcnp_version=DEFAULT_VERSION):
    """
    Gets the maximum allowed length for an input line for a specific MCNP version.

    The version must be a three component tuple e.g., (6, 2, 0) and (5, 1, 60).
    Line lengths inferred from:

    | C. J. Werner et al., “MCNP Version 6.2 Release Notes,” LA-UR--18-20808, 1419730, Feb. 2018. doi: 10.2172/1419730. section 2.6.2

    Prior MCNP release version numbers were taken from `RSICC <https://rsicc.ornl.gov/codes/ccc/ccc8/ccc-810.html>`_.

    :param mcnp_version: The version of MCNP that the input is intended for.
    :type mcnp_version: tuple
    :returns: The number of characters allowed in a line.
    :rtype: int
    """
    if mcnp_version >= DEFAULT_VERSION:
        return LINE_LENGTH[DEFAULT_VERSION]
    try:
        return LINE_LENGTH[mcnp_version]
    except KeyError:
        raise UnsupportedFeature(f"MCNP Version {mcnp_version} is not supported.")



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
class LineOverRunWarning(UserWarning):
    """
    Raised when non-comment inputs exceed the allowed line length in an input.
    """

    def __init__(self, message):
        self.message = message


class MalformedInputError(ValueError):
    """
    Raised when there is an error with the MCNP input not related to the parser.
    """

    def __init__(self, input, message):
        if input and getattr(input, "input_file", None) and input.input_file:
            self.input = input
            path = input.input_file.path
            start_line = input.line_number
            self.path = path
            self.start = start_line
            lines = "\n".join(input.input_lines)
        else:
            path = ""
            start_line = 0
            lines = ""
        self.message = f"""
: {path}, line {start_line}
{message}

the full input:
{lines}"""
        super().__init__(self.message)


class ParsingError(MalformedInputError):
    """
    Raised when there is an error parsing the MCNP input at the SLY parsing layer.
    """

    def __init__(self, input, message, error_queue):
        messages = []
        if input and getattr(input, "input_file", None) and input.input_file:
            self.input = input
            path = input.input_file.path
            start_line = input.line_number
            self.path = path
            self.start = start_line
        else:
            path = ""
            start_line = 0
        if error_queue:
            for error in error_queue:
                if token := error["token"]:
                    line_no = error["line"]
                    index = error["index"]
                    self.rel_line = line_no
                    self.abs_line = line_no + start_line
                    base_message = f'There was an error parsing "{token.value}".'
                else:
                    line_no = 0
                    index = 0
                    base_message = f"The input ended prematurely."
                buffer = [f"    {path}, line {start_line + line_no -1}", ""]
                if input:
                    for i, line in enumerate(input.input_lines):
                        if i == line_no - 1:
                            buffer.append(f"    >{start_line + i:5g}| {line}")
                            if token:
                                length = len(token.value)
                                marker = "^" * length
                                buffer.append(
                                    f"{' '* 10}|{' ' * (index+1)}{marker} not expected here."
                                )
                        else:
                            buffer.append(f"     {start_line + i:5g}| {line}")
                    buffer.append(base_message)
                    buffer.append(error["message"])
                messages.append("\n".join(buffer))
            self.message = "\n".join(messages + [message])
        else:
            self.message = message

        ValueError.__init__(self, self.message)


class NumberConflictError(Exception):
    """
    Raised when there is a conflict in number spaces
    """

    def __init__(self, message):
        self.message = message
        super().__init__(message)


class BrokenObjectLinkError(MalformedInputError):
    """Raised when the referenced object does not exist in the input file."""

    def __init__(self, parent_type, parent_number, child_type, child_number):
        """
        :param parent_type: Name of the parent object linking (e.g., Cell)
        :type parent_type: str
        :param parent_number: the number of the parent object
        :type parent_number: int
        :param child_type: Name of the type of object missing in the link (e.g., surface)
        :type child_type: str
        :param child_number: the number for the missing object
        :type child_number: int
        """
        super().__init__(
            None,
            f"{child_type} {child_number} is missing from the input from the definition of: {parent_type} {parent_number}",
        )


class RedundantParameterSpecification(ValueError):
    """
    Raised when multiple conflicting parameters are given.

    e.g., ``1 0 -1 imp:n=5 imp:n=0``
    """

    def __init__(self, key, new_value):
        self.message = (
            f"Multiple values given for parameter: {key}. new_value given: {new_value}."
        )
        super().__init__(self.message)


class ParticleTypeNotInProblem(ValueError):
    """
    Raised when data are set for a particle type not in
    the problem's mode.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(message)


class ParticleTypeNotInCell(ValueError):
    """
    Raised when data for importance data for a particle in
    the problem is not provided for a cell.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(message)


class UnsupportedFeature(NotImplementedError):
    """
    Raised when MCNP syntax that is not supported is found
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class UnknownElement(ValueError):
    """
    Raised when an undefined element is used.
    """

    def __init__(self, missing_val):
        self.message = f"An element identified by: {missing_val} is unknown to MontePy."
        super().__init__(self.message)


class IllegalState(ValueError):
    """
    Raised when an object can't be printed out due to an illegal state.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class LineExpansionWarning(Warning):
    """
    Warning for when a field or line expands that may damage user formatting.
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import itertools
from montepy.data_inputs.cell_modifier import CellModifierInput
from montepy.errors import *
from montepy.constants import DEFAULT_VERSION
from montepy.input_parser.mcnp_input import Jump
from montepy.input_parser import syntax_node
from montepy.mcnp_object import MCNP_Object
from montepy.universe import Universe
from montepy.utilities import *


class UniverseInput(CellModifierInput):
    """
    Object to actually handle the ``U`` input in cells
    and data blocks.

    :param input: the Input object representing this data input
    :type input: Input
    :param in_cell_block: if this card came from the cell block of an input file.
    :type in_cell_block: bool
    :param key: the key from the key-value pair in a cell
    :type key: str
    :param value: the value syntax tree from the key-value pair in a cell
    :type value: SyntaxNode
    """

    def __init__(self, input=None, in_cell_block=False, key=None, value=None):
        self._universe = None
        self._old_numbers = []
        self._old_number = self._generate_default_node(int, Jump())
        self._not_truncated = False
        super().__init__(input, in_cell_block, key, value)
        if self.in_cell_block:
            if key:
                val = self._tree["data"][0]
                val.is_negatable_identifier = True
                self._not_truncated = val.is_negative
                self._old_number = val
        elif input:
            self._universes = []
            for node in self.data:
                try:
                    node.is_negatable_identifier = True
                    if node.value is not None:
                        self._old_numbers.append(node)
                    else:
                        self._old_numbers.append(node)
                except ValueError:
                    raise MalformedInputError(
                        input,
                        f"Cell universes must be an integer ≥ 0. {node} was given",
                    )

    def _generate_default_cell_tree(self):
        list_node = syntax_node.ListNode("number sequence")
        list_node.append(self._generate_default_node(int, Jump()))
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = self._generate_default_node(
            str, self._class_prefix().upper(), None
        )
        self._tree = syntax_node.SyntaxNode(
            "universe",
            {
                "classifier": classifier,
                "param_seperator": self._generate_default_node(str, "=", None),
                "data": list_node,
            },
        )

    @staticmethod
    def _class_prefix():
        return "u"

    @staticmethod
    def _has_number():
        return False

    @staticmethod
    def _has_classifier():
        return 0

    @make_prop_val_node("_old_number")
    def old_number(self):
        pass

    @property
    def old_numbers(self):
        ret = []
        for value in self._old_numbers:
            if isinstance(value, syntax_node.ValueNode):
                ret.append(value.value)
            else:
                ret.append(value)
        return ret

    @property
    def has_information(self):
        if self.in_cell_block:
            return self.universe is not None and self.universe.number != 0

    @make_prop_pointer("_universe", Universe)
    def universe(self):
        if self.in_cell_block:
            return self._universe

    @property
    def not_truncated(self):
        """
        Indicates if this cell has been marked as not being truncated for optimization.

        See Note 1 from section 3.3.1.5.1 of the user manual (LA-UR-17-29981).

            Note 1. A problem will run faster by preceding the U card entry with a minus sign for any
            cell that is not truncated by the boundary of any higher-level cell. (The minus sign indicates
            that calculating distances to boundary in higher-level cells can be omitted.) Use this
            capability with EXTREME CAUTION; MCNP6 cannot detect errors in this feature because
            the logic that enables detection is omitted by the presence of the negative universe. Extremely
            wrong answers can be quietly calculated. Plot several views of the geometry or run with the
            VOID card to check for errors.

            -- LA-UR-17-29981.

        :rtype: bool
        :returns: True if this cell has been marked as not being truncated by the parent filled cell.
        """
        return self._not_truncated

    @not_truncated.setter
    def not_truncated(self, value):
        if not isinstance(value, bool):
            raise TypeError("truncated_by_parent must be a bool")
        self._not_truncated = value

    @property
    def _tree_value(self):
        val = self._old_number
        val.value = self.universe.number
        val.is_negative = self.not_truncated
        return val

    def _collect_new_values(self):
        ret = []
        for cell in self._problem.cells:
            # force value update here with _tree_value
            if cell._universe._tree_value.value == 0:
                # access ValueNode directly to avoid override with _tree_value
                cell._universe._old_number._value = None
            ret.append(cell._universe._old_number)
        return ret

    def merge(self, other):
        raise MalformedInputError(
            other._input, "Cannot have two universe inputs for the problem"
        )

    def push_to_cells(self):
        if not self._problem:
            return
        if not self.in_cell_block:
            cells = self._problem.cells
            if self._old_numbers:
                self._check_redundant_definitions()
                for cell, uni_number in itertools.zip_longest(
                    cells, self._old_numbers, fillvalue=None
                ):
                    if isinstance(uni_number, (Jump, type(None))):
                        continue
                    cell._universe._old_number = uni_number
                    if uni_number.is_negative:
                        cell._universe._not_truncated = True
            universes = self._problem.universes
            for cell in cells:
                uni_num = cell.old_universe_number
                if uni_num is None:
                    uni_num = 0
                if uni_num not in universes.numbers:
                    universe = Universe(uni_num)
                    universe.link_to_problem(self._problem)
                    universes.append(universe)
                else:
                    universe = universes[uni_num]
                cell._universe._universe = universe

    def _clear_data(self):
        del self._old_numbers

    def __str__(self):
        ret = "\n".join(self.format_for_mcnp_input(DEFAULT_VERSION))
        return ret

    def __repr__(self):
        ret = (
            f"UNIVERSE: in_cell: {self._in_cell_block}"
            f" set_in_block: {self.set_in_cell_block}, "
            f"Universe : {self._universe}, "
            f"Old Numbers: {self._old_numbers}"
        )
        return ret

    def _update_cell_values(self):
        if self.universe is not None:
            self._tree["data"][0].is_negatable_identifier = True
            self._tree["data"][0].value = self.universe.number
            self._tree["data"][0].is_negative = self.not_truncated



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from unittest import TestCase
import montepy
from montepy.cell import Cell
from montepy.particle import Particle
from montepy.data_inputs.importance import Importance
from montepy.errors import *
from montepy.input_parser import mcnp_input, block_type
import os


class TestImportance(TestCase):
    def test_importance_init_cell(self):
        # test_normal cell init
        in_str = "1 0 -1 IMP:N,P=1"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        cell = Cell(card)
        self.assertEqual(cell.importance.neutron, 1.0)
        self.assertEqual(cell.importance.photon, 1.0)
        self.assertEqual(cell.importance.alpha_particle, 0.0)
        self.assertIsNone(cell.importance.all)
        self.assertTrue(cell.importance.in_cell_block)
        # test non-number imp
        in_str = "1 0 -1 IMP:N,P=h"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        with self.assertRaises(ValueError):
            cell = Cell(card)
        # test negative imp
        in_str = "1 0 -1 IMP:N,P=-2"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        with self.assertRaises(ValueError):
            cell = Cell(card)

    def test_importance_init_data(self):
        in_str = "IMP:N,P 1 0"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        imp = Importance(card)
        self.assertEqual(
            [val.value for val in imp._particle_importances[Particle.NEUTRON]["data"]],
            [1.0, 0.0],
        )
        self.assertEqual(
            [val.value for val in imp._particle_importances[Particle.PHOTON]["data"]],
            [1.0, 0.0],
        )
        # test non-number imp
        in_str = "IMP:N,P 1 h"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        with self.assertRaises(ValueError):
            imp = Importance(card)
        # test negative
        in_str = "IMP:N,P 1 -2"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        with self.assertRaises(ValueError):
            imp = Importance(card)
        # test bad in_cell_block
        in_str = "IMP:N,P 1 2"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        with self.assertRaises(TypeError):
            imp = Importance(card, in_cell_block=1)
        # test bad key
        in_str = "IMP:N,P 1 2"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        with self.assertRaises(TypeError):
            imp = Importance(card, key=1)
        # test bad value
        in_str = "IMP:N,P 1 2"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        with self.assertRaises(TypeError):
            imp = Importance(card, value=1)

    def test_importance_iter_getter_in(self):
        in_str = "1 0 -1 IMP:N,P=1"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        cell = Cell(card)
        imp = cell.importance
        particles = [
            montepy.particle.Particle.NEUTRON,
            montepy.particle.Particle.PHOTON,
        ]
        for particle in imp:
            self.assertIn(particle, particles)
            self.assertAlmostEqual(imp[particle], 1.0)
        for particle in particles:
            self.assertIn(particle, imp)
        with self.assertRaises(TypeError):
            imp["hi"]

    def test_importance_all_setter(self):
        in_str = "1 0 -1 IMP:N,P=1"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        cell = Cell(card)
        problem = montepy.mcnp_problem.MCNP_Problem("foo")
        problem.mode.add(montepy.particle.Particle.NEUTRON)
        problem.mode.add(montepy.particle.Particle.PHOTON)
        imp = cell.importance
        cell.link_to_problem(problem)
        imp.all = 2.0
        self.assertAlmostEqual(imp.neutron, 2.0)
        self.assertAlmostEqual(imp.photon, 2.0)
        # try wrong type
        with self.assertRaises(TypeError):
            imp.all = "h"
        # try negative type
        with self.assertRaises(ValueError):
            imp.all = -2.0

    def test_importance_setter(self):
        in_str = "1 0 -1 IMP:N,P=1"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        cell = Cell(card)
        cell.importance.neutron = 2.5
        self.assertEqual(cell.importance.neutron, 2.5)
        problem = montepy.mcnp_problem.MCNP_Problem("foo")
        cell.link_to_problem(problem)
        # test problem mode enforcement
        with self.assertRaises(ValueError):
            cell.importance.photon = 1.0
        # test wrong type
        with self.assertRaises(TypeError):
            cell.importance.neutron = "h"
        # test negative
        with self.assertRaises(ValueError):
            cell.importance.neutron = -0.5

        cell.importance[Particle.NEUTRON] = 3
        self.assertEqual(cell.importance.neutron, 3.0)
        with self.assertRaises(TypeError):
            cell.importance[""] = 5
        with self.assertRaises(TypeError):
            cell.importance[Particle.NEUTRON] = ""
        with self.assertRaises(ValueError):
            cell.importance[Particle.NEUTRON] = -1.0

    def test_importance_deleter(self):
        in_str = "1 0 -1 IMP:N,P=1"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        cell = Cell(card)
        del cell.importance.neutron
        self.assertAlmostEqual(cell.importance.neutron, 0.0)
        del cell.importance[Particle.PHOTON]
        self.assertAlmostEqual(cell.importance.photon, 0.0)
        with self.assertRaises(TypeError):
            del cell.importance[""]

    def test_importance_merge(self):
        in_str = "IMP:N,P 1 0"
        card = mcnp_input.Input([in_str], block_type.BlockType.DATA)
        imp1 = Importance(card)
        in_str = "IMP:E 0 0"
        card = mcnp_input.Input([in_str], block_type.BlockType.DATA)
        imp2 = Importance(card)
        imp1.merge(imp2)
        self.assertEqual(
            [val.value for val in imp1._particle_importances[Particle.NEUTRON]["data"]],
            [1.0, 0.0],
        )
        self.assertEqual(
            [
                val.value
                for val in imp1._particle_importances[Particle.ELECTRON]["data"]
            ],
            [0.0, 0.0],
        )
        # test bad type
        with self.assertRaises(TypeError):
            imp1.merge("hi")
        # test bad block type
        in_str = "1 0 -1 IMP:N,P=1"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        cell = Cell(card)
        with self.assertRaises(ValueError):
            imp1.merge(cell.importance)
        in_str = "IMP:P 0 0"
        card = mcnp_input.Input([in_str], block_type.BlockType.CELL)
        imp2 = Importance(card)
        with self.assertRaises(MalformedInputError):
            imp1.merge(imp2)

    def tests_redundant_importance(self):
        with self.assertRaises(MalformedInputError):
            montepy.read_input(
                os.path.join("tests", "inputs", "test_imp_redundant.imcnp")
            )



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import copy
from montepy.errors import *
from montepy.data_inputs import transform
from montepy.input_parser import syntax_node
from montepy.input_parser.surface_parser import SurfaceParser
from montepy.numbered_mcnp_object import Numbered_MCNP_Object
from montepy.surfaces import half_space
from montepy.surfaces.surface_type import SurfaceType
from montepy.utilities import *
import re


def _enforce_numbers(self, value):
    if value <= 0:
        raise ValueError(f"The number be greater than 0; {value} given.")


class Surface(Numbered_MCNP_Object):
    """
    Object to hold a single MCNP surface

    :param input: The Input object representing the input
    :type input: Input
    """

    _parser = SurfaceParser()

    def __init__(self, input=None):
        super().__init__(input, self._parser)
        self._periodic_surface = None
        self._old_periodic_surface = self._generate_default_node(int, None)
        self._transform = None
        self._old_transform_number = self._generate_default_node(int, None)
        self._is_reflecting = False
        self._is_white_boundary = False
        self._surface_constants = []
        self._surface_type = self._generate_default_node(str, None)
        self._number = self._generate_default_node(int, -1)
        self._modifier = self._generate_default_node(str, None)
        # surface number
        if input:
            self._number = self._tree["surface_num"]["number"]
            self._number._convert_to_int()
            self._old_number = copy.deepcopy(self._number)
            if "modifier" in self._tree["surface_num"]:
                self._modifier = self._tree["surface_num"]["modifier"]
                if self._modifier.value == "*":
                    self._is_reflecting = True
                elif "+" in self._number.token:
                    self._is_white_boundary = True
                    self._number._token = self._number.token.replace("+", "")
                    self._modifier = self._generate_default_node(str, "+", None)
                    self._tree["surface_num"].nodes["modifier"] = self._modifier
            try:
                assert self._number.value > 0
            except AssertionError:
                raise MalformedInputError(
                    input,
                    f"{self._number.value} could not be parsed as a valid surface number.",
                )
            if self._tree["pointer"].value is not None:
                val = self._tree["pointer"]
                val.is_negatable_identifier = True
                if val.is_negative:
                    self._old_periodic_surface = val
                else:
                    self._old_transform_number = val
            self._surface_type = self._tree["surface_type"]
            # parse surface mnemonic
            try:
                # enforce enums
                self._surface_type._convert_to_enum(SurfaceType, switch_to_upper=True)
            # this should never be reached due to SLY rules.
            # still if it is somehow reached this error is more helpful to the user.
            except ValueError:  # pragma: no cover
                raise MalformedInputError(
                    input,
                    f"{self._surface_type.value} could not be parsed as a surface type mnemonic.",
                )
            # parse the parameters
            for entry in self._tree["data"]:
                self._surface_constants.append(entry)

    @make_prop_val_node("_surface_type", (SurfaceType, str), SurfaceType)
    def surface_type(self):
        """
        The mnemonic for the type of surface.

        E.g. CY, PX, etc.

        :rtype: SurfaceType
        """
        pass

    @property
    def is_reflecting(self):
        """
        If true this surface is a reflecting boundary.

        :rtype: bool
        """
        return self._is_reflecting

    @is_reflecting.setter
    def is_reflecting(self, reflect):
        if not isinstance(reflect, bool):
            raise TypeError("is_reflecting must be set to a bool")
        self._is_reflecting = reflect

    @property
    def is_white_boundary(self):
        """
        If true this surface is a white boundary.

        :rtype: bool
        """
        return self._is_white_boundary

    @is_white_boundary.setter
    def is_white_boundary(self, white):
        if not isinstance(white, bool):
            raise TypeError("is_white_boundary must be set to a bool")
        self._is_white_boundary = white

    @property
    def surface_constants(self):
        """
        The constants defining the surface

        :rtype: list
        """
        ret = []
        for val in self._surface_constants:
            ret.append(val.value)
        return ret

    @surface_constants.setter
    def surface_constants(self, constants):
        if not isinstance(constants, list):
            raise TypeError("surface_constants must be a list")
        if len(constants) != len(self._surface_constants):
            raise ValueError(f"Cannot change the length of the surface constants.")
        for constant in constants:
            if not isinstance(constant, float):
                raise TypeError(
                    f"The surface constant provided: {constant} must be a float"
                )
        for i, value in enumerate(constants):
            self._surface_constants[i].value = value

    @make_prop_val_node("_old_transform_number")
    def old_transform_number(self):
        """
        The transformation number for this surface in the original file.

        :rtype: int
        """
        pass

    @make_prop_val_node("_old_periodic_surface")
    def old_periodic_surface(self):
        """
        The surface number this is periodic with reference to in the original file.

        :rtype: int
        """
        pass

    @make_prop_pointer("_periodic_surface", types=(), deletable=True)
    def periodic_surface(self):
        """
        The surface that this surface is periodic with respect to

        :rtype: Surface
        """
        pass

    @make_prop_pointer("_transform", transform.Transform, deletable=True)
    def transform(self):
        """
        The Transform object that translates this surface

        :rtype: Transform
        """
        pass

    @make_prop_val_node("_old_number")
    def old_number(self):
        """
        The surface number that was used in the read file

        :rtype: int
        """
        pass

    @make_prop_val_node("_number", int, validator=_enforce_numbers)
    def number(self):
        """
        The surface number to use.

        :rtype: int
        """
        pass

    @property
    def cells(self):
        """
        A generator of Cells that use this surface.

        :rtype: generator
        """
        if self._problem:
            for cell in self._problem.cells:
                if self in cell.surfaces:
                    yield cell

    def __str__(self):
        return f"SURFACE: {self.number}, {self.surface_type}"

    def __repr__(self):
        return (
            f"SURFACE: {self.number}, {self.surface_type}, "
            f"periodic surface: {self.periodic_surface}, "
            f"transform: {self.transform}, "
            f"constants: {self.surface_constants}"
        )

    def update_pointers(self, surfaces, data_inputs):
        """
        Updates the internal pointers to the appropriate objects.

        Right now only periodic surface links will be made.
        Eventually transform pointers should be made.

        :param surfaces: A Surfaces collection of the surfaces in the problem.
        :type surfaces: Surfaces
        :param data_cards: the data_cards in the problem.
        :type data_cards: list
        """
        if self.old_periodic_surface:
            try:
                self._periodic_surface = surfaces[self.old_periodic_surface]
            except KeyError:
                raise BrokenObjectLinkError(
                    "Surface",
                    self.number,
                    "Periodic Surface",
                    self.old_periodic_surface,
                )
        if self.old_transform_number:
            for input in data_inputs:
                if isinstance(input, transform.Transform):
                    if input.number == self.old_transform_number:
                        self._transform = input
            if not self.transform:
                raise BrokenObjectLinkError(
                    "Surface",
                    self.number,
                    "Transform",
                    self.old_transform_number,
                )

    def validate(self):
        if self.surface_type is None:
            raise IllegalState(
                f"Surface: {self.number} does not have a surface type set."
            )

    def _update_values(self):
        modifier = self._tree["surface_num"]["modifier"]
        if self.is_reflecting:
            modifier.value = "*"
        elif self.is_white_boundary:
            modifier.value = "+"
        if self.transform is not None:
            self._old_transform_number.value = self.transform.number
            self._old_transform_number.is_negative = False
            self._tree.nodes["pointer"] = self._old_transform_number
        elif self.periodic_surface is not None:
            self._old_periodic_surface.value = self.periodic_surface.number
            self._old_periodic_surface.is_negative = True
            self._tree.nodes["pointer"] = self._old_periodic_surface

    def __lt__(self, other):
        return self.number < other.number

    def __eq__(self, other):
        return (
            self.number == other.number
            and self.surface_type == other.surface_type
            and self.is_reflecting == other.is_reflecting
            and self.is_white_boundary == other.is_white_boundary
            and self.surface_constants == other.surface_constants
        )

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((self.number, str(self.surface_type)))

    def __eq__(self, other):
        return (
            self.number == other.number
            and self.surface_type == other.surface_type
            and self.is_reflecting == other.is_reflecting
            and self.is_white_boundary == other.is_white_boundary
            and self.surface_constants == other.surface_constants
        )

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((self.number, str(self.surface_type)))

    def find_duplicate_surfaces(self, surfaces, tolerance):
        """Finds all surfaces that are effectively the same as this one.

        :param surfaces: a list of the surfaces to compare against this one.
        :type surfaces: list
        :param tolerance: the amount of relative error to allow
        :type tolerance: float

        :returns: A list of the surfaces that are identical
        :rtype: list
        """
        return []

    def __neg__(self):
        return half_space.UnitHalfSpace(self, False, False)

    def __pos__(self):
        return half_space.UnitHalfSpace(self, True, False)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from unittest import TestCase

import montepy

from montepy._cell_data_control import CellDataPrintController
from montepy.data_inputs.data_input import DataInput
from montepy.data_inputs import material, thermal_scattering, transform, volume
from montepy.data_inputs.data_parser import parse_data
from montepy.errors import *
from montepy.input_parser.mcnp_input import Input, Jump
from montepy.input_parser import syntax_node
from montepy.input_parser.block_type import BlockType
from montepy.mcnp_problem import MCNP_Problem


class testDataInputClass(TestCase):
    def test_data_card_init(self):
        in_str = "imp:n 1 1 1 0"
        input_card = Input([in_str], BlockType.DATA)
        data_card = DataInput(input_card)
        words = in_str.split()[1:]
        for i, word in enumerate(data_card.data):
            self.assertEqual(word.value, float(words[i]))
        # TODO test comments
        # self.assertEqual(comment, data_card.comments[0])

    def test_data_card_empty_constructor(self):
        card = DataInput()

    def test_data_card_str(self):
        in_str = "vol 1 1 0"
        input_card = Input([in_str], BlockType.DATA)
        data_card = DataInput(input_card)
        self.assertEqual(str(data_card), "DATA INPUT: vol ")

    def test_data_card_format_mcnp(self):
        in_strs = ["c foo", "c bar", "m1 1001.80c 1.0 $ bar"]
        input_card = Input(in_strs, BlockType.DATA)
        data_card = DataInput(input_card)
        output = data_card.format_for_mcnp_input((6, 2, 0))
        print(output)
        self.assertEqual(len(output), len(in_strs))
        for answer, out in zip(in_strs, output):
            self.assertEqual(answer, out)

    # TODO implement comment setting
    def test_comment_setter(self):
        in_str = "m1 1001.80c 1.0"
        input_card = Input([in_str], BlockType.DATA)
        comment = "foo"
        data_card = DataInput(input_card)
        data_card.comment = comment
        self.assertEqual(comment, data_card.comment)

    def test_data_parser(self):
        identifiers = {
            "m235": material.Material,
            "mt235": thermal_scattering.ThermalScatteringLaw,
            "tr601": transform.Transform,
            "ksrc": DataInput,
        }
        in_strs = {
            "m235": "m235 1001.80c 1.0",
            "mt235": "mt235 grph.29t",
            "tr601": "tr601 0.0 0.0 10.",
            "ksrc": "ksrc 1.0 0.0 0.0",
        }

        for identifier, w in in_strs.items():
            for ident in [identifier, identifier.upper()]:
                input_card = Input([w], BlockType.DATA)
                card = parse_data(input_card)
                self.assertIsInstance(card, identifiers[ident.lower()])

    def test_data_card_mutate_print(self):
        in_str = "IMP:N 1 1"
        input_card = Input([in_str], BlockType.DATA)
        input_card = DataInput(input_card)
        input_card._mutated = True
        output = input_card.format_for_mcnp_input((6, 2, 0))
        self.assertEqual(output, [in_str])

    def test_print_in_data_block(self):
        cell_controller = CellDataPrintController()
        cell_controller["imp"] = True
        cell_controller["Imp"] = True
        self.assertTrue(cell_controller["IMP"])
        with self.assertRaises(TypeError):
            cell_controller[5] = True
        with self.assertRaises(TypeError):
            cell_controller["imp"] = 5
        with self.assertRaises(TypeError):
            cell_controller[5]
        with self.assertRaises(KeyError):
            cell_controller["a"] = True
        with self.assertRaises(KeyError):
            cell_controller["a"]

    def test_volume_init_cell(self):
        vol = 1.0
        list_node = syntax_node.ListNode("data")
        list_node.append(syntax_node.ValueNode(str(vol), float))
        node = syntax_node.SyntaxNode(
            "volume",
            {
                "classifier": syntax_node.ValueNode("VoL", str),
                "seperator": syntax_node.ValueNode("=", str),
                "data": list_node,
            },
        )
        card = volume.Volume(key="VoL", value=node, in_cell_block=True)
        self.assertEqual(card.volume, vol)
        self.assertTrue(card.in_cell_block)
        self.assertTrue(card.set_in_cell_block)
        self.assertTrue(not card.is_mcnp_calculated)
        self.assertTrue(card.set)
        card = volume.Volume(in_cell_block=True)
        self.assertTrue(card.is_mcnp_calculated)
        with self.assertRaises(ValueError):
            list_node = syntax_node.ListNode("data")
            list_node.append(syntax_node.ValueNode("s", str))
            node = syntax_node.SyntaxNode(
                "volume",
                {
                    "classifier": syntax_node.ValueNode("VoL", str),
                    "seperator": syntax_node.ValueNode("=", str),
                    "data": list_node,
                },
            )
            card = volume.Volume(key="VoL", value=node, in_cell_block=True)
        with self.assertRaises(ValueError):
            list_node = syntax_node.ListNode("data")
            list_node.append(syntax_node.ValueNode("-1", float))
            node = syntax_node.SyntaxNode(
                "volume",
                {
                    "classifier": syntax_node.ValueNode("VoL", str),
                    "seperator": syntax_node.ValueNode("=", str),
                    "data": list_node,
                },
            )
            card = volume.Volume(key="VoL", value=node, in_cell_block=True)

    # TODO merge manually
    """
    def test_volume_init_data(self):
        in_str = "VOL 1 1 2J 0"
        input_card = Card([in_str], BlockType.DATA)
        vol_card = parse_data(input_card)
        answers = [1.0, 1.0, Jump, Jump, 0.0]
        for i, vol in enumerate(vol_card._volume):
            if isinstance(answers[i], float):
                self.assertAlmostEqual(vol, answers[i])
            else:
                self.assertIsInstance(vol, Jump)
        # test starting jump
        in_str = "VOL 2J 1 0"
        input_card = Card([in_str], BlockType.DATA)
        vol_card = parse_data(input_card)
        answers = [Jump, Jump, 1.0, 0.0]
        for i, vol in enumerate(vol_card._volume):
            if isinstance(answers[i], float):
                self.assertAlmostEqual(vol, answers[i])
            else:
                self.assertIsInstance(vol, Jump)
        # tests starting no
        in_str = "VOL NO 1 1 2J 0"
        input_card = Card([in_str], BlockType.DATA)
        vol_card = parse_data(input_card)
        self.assertTrue(not vol_card.is_mcnp_calculated)
        # invalid number
        in_str = "VOL NO s 1 2J 0"
        input_card = Card([in_str], BlockType.DATA)
        with self.assertRaises(MalformedInputError):
            vol_card = parse_data(input_card)
        # negative volume
        in_str = "VOL NO -1 1 2J 0"
        input_card = Card([in_str], BlockType.DATA)
        with self.assertRaises(MalformedInputError):
            vol_card = parse_data(input_card)
    """

    def test_volume_init_data(self):
        in_str = "VOL 1 1 2J 0"
        input_card = Input([in_str], BlockType.DATA)
        vol_card = parse_data(input_card)
        answers = [1.0, 1.0, None, None, 0.0]
        for i, vol in enumerate(vol_card._volume):
            if isinstance(vol, syntax_node.ValueNode):
                self.assertAlmostEqual(vol.value, answers[i])
            else:
                self.assertIsInstance(vol, Jump)
        in_str = "VOL NO 1 1 2J 0"
        input_card = Input([in_str], BlockType.DATA)
        vol_card = parse_data(input_card)
        self.assertTrue(not vol_card.is_mcnp_calculated)
        # invalid number
        in_str = "VOL NO s 1 2J 0"
        input_card = Input([in_str], BlockType.DATA)
        with self.assertRaises(MalformedInputError):
            vol_card = parse_data(input_card)
        # negative volume
        in_str = "VOL NO -1 1 2J 0"
        input_card = Input([in_str], BlockType.DATA)
        with self.assertRaises(MalformedInputError):
            vol_card = parse_data(input_card)

    def test_volumes_for_only_some_cells(self):
        cells = [
            montepy.Cell(Input([f"{i + 1} 0 -1 u=3"], BlockType.CELL))
            for i in range(10)
        ]
        prob = MCNP_Problem(None)
        prob.cells = cells
        vol_card = Input(["VOL 1 1 2 3 5"], BlockType.DATA)
        vol_data = volume.Volume(vol_card, in_cell_block=False)
        vol_data.link_to_problem(prob)
        vol_data.push_to_cells()
        is_set = [cell.volume_is_set for cell in cells]
        reference = [True] * 5 + [False] * 5
        self.assertListEqual(is_set, reference)

    def test_volume_setter(self):
        vol = 1.0
        list_node = syntax_node.ListNode("data")
        list_node.append(syntax_node.ValueNode(str(vol), float))
        node = syntax_node.SyntaxNode(
            "volume",
            {
                "classifier": syntax_node.ValueNode("VoL", str),
                "seperator": syntax_node.ValueNode("=", str),
                "data": list_node,
            },
        )
        card = volume.Volume(key="VoL", value=node, in_cell_block=True)
        card.volume = 5.0
        self.assertEqual(card.volume, 5.0)
        with self.assertRaises(TypeError):
            card.volume = "hi"
        with self.assertRaises(ValueError):
            card.volume = -5.0

    def test_volume_deleter(self):
        vol = 1.0
        list_node = syntax_node.ListNode("data")
        list_node.append(syntax_node.ValueNode(str(vol), float))
        node = syntax_node.SyntaxNode(
            "volume",
            {
                "classifier": syntax_node.ValueNode("VoL", str),
                "seperator": syntax_node.ValueNode("=", str),
                "data": list_node,
            },
        )
        card = volume.Volume(key="VoL", value=node, in_cell_block=True)
        del card.volume
        self.assertIsNone(card.volume)

    def test_volume_merge(self):
        vol = 1.0
        list_node = syntax_node.ListNode("data")
        list_node.append(syntax_node.ValueNode(str(vol), float))
        node = syntax_node.SyntaxNode(
            "volume",
            {
                "classifier": syntax_node.ValueNode("VoL", str),
                "seperator": syntax_node.ValueNode("=", str),
                "data": list_node,
            },
        )
        card = volume.Volume(key="VoL", value=node, in_cell_block=True)
        card2 = volume.Volume(key="VoL", value=node, in_cell_block=True)
        with self.assertRaises(MalformedInputError):
            card.merge(card2)

    def test_volume_repr(self):
        vol = 1.0
        list_node = syntax_node.ListNode("data")
        list_node.append(syntax_node.ValueNode(str(vol), float))
        node = syntax_node.SyntaxNode(
            "volume",
            {
                "classifier": syntax_node.ValueNode("VoL", str),
                "seperator": syntax_node.ValueNode("=", str),
                "data": list_node,
            },
        )
        card = volume.Volume(key="VoL", value=node, in_cell_block=True)
        self.assertIn("VOLUME", repr(card))


class TestClassifier(TestCase):
    def test_classifier_start_comment(self):
        lines = [
            "c ******************************************************************************",
            "c *******************************Data Cards*************************************",
            "c ******************************************************************************",
            "c",
            "c",
            "c data file=fuel",
            "c ******************** Begin Fuel Data *****************************************",
            "c",
            "c",
            "c fuel",
            "c     material number 2001 total atom density =  ",
            "m2001",
            "      92235.50c     5.0E-01",
        ]
        input = Input(lines, BlockType.DATA)
        data = parse_data(input)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import copy
import unittest
from unittest import TestCase, expectedFailure
import os

import montepy
from montepy.data_inputs import material, thermal_scattering, volume
from montepy.input_parser.mcnp_input import (
    Input,
    Jump,
    Message,
    Title,
    ReadInput,
)
from montepy.errors import *
from montepy.particle import Particle
import numpy as np


class testFullFileIntegration(TestCase):
    @classmethod
    def setUpClass(cls):
        file_name = "tests/inputs/test.imcnp"
        cls.simple_problem = montepy.read_input(file_name)
        cls.importance_problem = montepy.read_input(
            os.path.join("tests", "inputs", "test_importance.imcnp")
        )
        cls.universe_problem = montepy.read_input(
            os.path.join("tests", "inputs", "test_universe.imcnp")
        )

    def test_original_input(self):
        cell_order = [Message, Title] + [Input] * 17
        for i, input_ob in enumerate(self.simple_problem.original_inputs):
            self.assertIsInstance(input_ob, cell_order[i])

    def test_original_input_dos(self):
        problem = montepy.read_input(os.path.join("tests", "inputs", "test_dos.imcnp"))
        cell_order = [Message, Title] + [Input] * 16
        for i, input_ob in enumerate(problem.original_inputs):
            self.assertIsInstance(input_ob, cell_order[i])

    def test_original_input_tabs(self):
        problem = montepy.read_input(os.path.join("tests", "inputs", "test_tab.imcnp"))
        cell_order = [Message, Title] + [Input] * 17
        for i, input_ob in enumerate(problem.original_inputs):
            self.assertIsInstance(input_ob, cell_order[i])

    # TODO formalize this or see if this is covered by other tests.
    def test_lazy_comments_check(self):
        problem = self.simple_problem
        material = problem.materials[2]
        for comment in material._tree.comments:
            print(repr(comment))
        print(material._tree.get_trailing_comment())

    def test_material_parsing(self):
        mat_numbers = [1, 2, 3]
        for i, mat in enumerate(self.simple_problem.materials):
            self.assertEqual(mat.number, mat_numbers[i])

    def test_surface_parsing(self):
        surf_numbers = [1000, 1005, 1010]
        for i, surf in enumerate(self.simple_problem.surfaces):
            self.assertEqual(surf.number, surf_numbers[i])

    def test_data_card_parsing(self):
        M = material.Material
        V = volume.Volume
        cards = [M, M, M, "KSRC", "KCODE", "PHYS:P", "MODE", V]
        for i, card in enumerate(self.simple_problem.data_inputs):
            if isinstance(cards[i], str):
                self.assertEqual(card.classifier.format().upper().rstrip(), cards[i])
            else:
                self.assertIsInstance(card, cards[i])
            if i == 2:
                self.assertTrue(card.thermal_scattering is not None)

    def test_cells_parsing_linking(self):
        cell_numbers = [1, 2, 3, 99, 5]
        mats = self.simple_problem.materials
        mat_answer = [mats[1], mats[2], mats[3], None, None]
        surfs = self.simple_problem.surfaces
        surf_answer = [{surfs[1000]}, {surfs[1005]}, set(surfs), {surfs[1010]}, set()]
        cells = self.simple_problem.cells
        complements = [set()] * 4 + [{cells[99]}]
        for i, cell in enumerate(self.simple_problem.cells):
            self.assertEqual(cell.number, cell_numbers[i])
            self.assertEqual(cell.material, mat_answer[i])
            surfaces = set(cell.surfaces)
            self.assertTrue(surfaces.union(surf_answer[i]) == surfaces)
            self.assertTrue(
                set(cell.complements).union(complements[i]) == complements[i]
            )

    def test_message(self):
        lines = ["this is a message", "it should show up at the beginning", "foo"]
        for i, line in enumerate(self.simple_problem.message.lines):
            self.assertEqual(line, lines[i])

    def test_title(self):
        answer = "MCNP Test Model for MOAA"
        self.assertEqual(answer, self.simple_problem.title.title)

    def test_read_card_recursion(self):
        problem = montepy.read_input("tests/inputs/testReadRec1.imcnp")
        self.assertEqual(len(problem.cells), 1)
        self.assertEqual(len(problem.surfaces), 1)
        self.assertIn(montepy.particle.Particle.PHOTON, problem.mode)

    def test_problem_str(self):
        output = str(self.simple_problem)
        self.assertIn("MCNP problem for: tests/inputs/test.imcnp", output)

    def test_write_to_file(self):
        out = "foo.imcnp"
        try:
            problem = copy.deepcopy(self.simple_problem)
            problem.write_to_file(out)
            with open(out, "r") as fh:
                for line in fh:
                    print(line.rstrip())
            test_problem = montepy.read_input(out)
            for i, cell in enumerate(self.simple_problem.cells):
                num = cell.number
                self.assertEqual(num, test_problem.cells[num].number)
                for attr in {"universe", "volume"}:
                    print(f"testing the attribute: {attr}")
                    gold = getattr(self.simple_problem.cells[num], attr)
                    test = getattr(test_problem.cells[num], attr)
                    if attr in {"universe"} and gold is not None:
                        gold, test = (gold.number, test.number)
                    self.assertEqual(gold, test)
            for i, surf in enumerate(self.simple_problem.surfaces):
                num = surf.number
                self.assertEqual(surf.number, test_problem.surfaces[num].number)
            for i, data in enumerate(self.simple_problem.data_inputs):
                if isinstance(data, material.Material):
                    self.assertEqual(data.number, test_problem.data_inputs[i].number)
                elif isinstance(data, volume.Volume):
                    self.assertEqual(str(data), str(test_problem.data_inputs[i]))
                else:
                    print("Rewritten data", data.data)
                    print("Original input data", test_problem.data_inputs[i].data)
                    self.assertEqual(data.data, test_problem.data_inputs[i].data)
        finally:
            if os.path.exists(out):
                os.remove(out)

    def test_cell_material_setter(self):
        cell = copy.deepcopy(self.simple_problem.cells[1])
        mat = self.simple_problem.materials[2]
        cell.material = mat
        self.assertEqual(cell.material, mat)
        cell.material = None
        self.assertIsNone(cell.material)
        with self.assertRaises(TypeError):
            cell.material = 5

    def test_problem_cells_setter(self):
        problem = copy.deepcopy(self.simple_problem)
        cells = copy.deepcopy(self.simple_problem.cells)
        cells.remove(cells[1])
        with self.assertRaises(TypeError):
            problem.cells = 5
        with self.assertRaises(TypeError):
            problem.cells = [5]
        with self.assertRaises(TypeError):
            problem.cells.append(5)
        problem.cells = cells
        self.assertEqual(problem.cells.objects, cells.objects)
        problem.cells = list(cells)
        self.assertEqual(problem.cells[2], cells[2])
        # test that cell modifiers are still there
        output = problem.cells._importance.format_for_mcnp_input((6, 2, 0))

    def test_problem_test_setter(self):
        problem = copy.deepcopy(self.simple_problem)
        sample_title = "This is a title"
        problem.title = sample_title
        self.assertEqual(problem.title.title, sample_title)
        with self.assertRaises(TypeError):
            problem.title = 5

    def test_problem_children_adder(self):
        problem = copy.deepcopy(self.simple_problem)
        BT = montepy.input_parser.block_type.BlockType
        in_str = "5 SO 5.0"
        card = montepy.input_parser.mcnp_input.Input([in_str], BT.SURFACE)
        surf = montepy.surfaces.surface_builder.surface_builder(card)
        in_str = "M5 6000.70c 1.0"
        card = montepy.input_parser.mcnp_input.Input([in_str], BT.SURFACE)
        mat = montepy.data_inputs.material.Material(card)
        in_str = "TR1 0 0 1"
        input = montepy.input_parser.mcnp_input.Input([in_str], BT.DATA)
        transform = montepy.data_inputs.transform.Transform(input)
        surf.transform = transform
        cell_num = 1000
        cell = montepy.Cell()
        cell.material = mat
        cell.geometry = -surf
        cell.mass_density = 1.0
        cell.number = cell_num
        cell.universe = problem.universes[350]
        problem.cells.append(cell)
        problem.add_cell_children_to_problem()
        self.assertIn(surf, problem.surfaces)
        self.assertIn(mat, problem.materials)
        self.assertIn(mat, problem.data_inputs)
        self.assertIn(transform, problem.transforms)
        self.assertIn(transform, problem.data_inputs)
        for cell_num in [1, cell_num]:
            print(cell_num)
            if cell_num == 1000:
                with self.assertWarns(LineExpansionWarning):
                    output = problem.cells[cell_num].format_for_mcnp_input((6, 2, 0))
            else:
                output = problem.cells[cell_num].format_for_mcnp_input((6, 2, 0))
            print(output)
            self.assertIn("U=350", "\n".join(output).upper())

    def test_problem_mcnp_version_setter(self):
        problem = copy.deepcopy(self.simple_problem)
        with self.assertRaises(ValueError):
            problem.mcnp_version = (4, 5, 3)
        problem.mcnp_version = (6, 2, 5)
        self.assertEqual(problem.mcnp_version, (6, 2, 5))

    def test_problem_duplicate_surface_remover(self):
        problem = montepy.read_input("tests/inputs/test_redundant_surf.imcnp")
        surfaces = problem.surfaces
        nums = list(problem.surfaces.numbers)
        survivors = (
            nums[0:3] + nums[5:8] + [nums[9]] + nums[11:13] + [nums[13]] + [nums[-2]]
        )
        problem.remove_duplicate_surfaces(1e-4)
        self.assertEqual(list(problem.surfaces.numbers), survivors)
        cell_surf_answer = "-1 3 -6"
        self.assertIn(
            cell_surf_answer, problem.cells[2].format_for_mcnp_input((6, 2, 0))[1]
        )

    def test_surface_periodic(self):
        problem = montepy.read_input("tests/inputs/test_surfaces.imcnp")
        surf = problem.surfaces[1]
        periodic = problem.surfaces[2]
        self.assertEqual(surf.periodic_surface, periodic)
        self.assertIn("1 -2 SO", surf.format_for_mcnp_input((6, 2, 0))[0])
        surf.periodic_surface = problem.surfaces[3]
        self.assertEqual(surf.periodic_surface, problem.surfaces[3])
        del surf.periodic_surface
        self.assertIsNone(surf.periodic_surface)
        with self.assertRaises(TypeError):
            surf.periodic_surface = 5

    def test_surface_transform(self):
        problem = montepy.read_input("tests/inputs/test_surfaces.imcnp")
        surf = problem.surfaces[1]
        transform = problem.data_inputs[0]
        del surf.periodic_surface
        surf.transform = transform
        self.assertEqual(surf.transform, transform)
        self.assertIn("1 1 SO", surf.format_for_mcnp_input((6, 2, 0))[0])
        del surf.transform
        self.assertIsNone(surf.transform)
        self.assertIn("1 SO", surf.format_for_mcnp_input((6, 2, 0))[0])
        with self.assertRaises(TypeError):
            surf.transform = 5

    def test_materials_setter(self):
        problem = copy.deepcopy(self.simple_problem)
        with self.assertRaises(TypeError):
            problem.materials = 5
        with self.assertRaises(TypeError):
            problem.materials = [5]
        size = len(problem.materials)
        problem.materials = list(problem.materials)
        self.assertEqual(len(problem.materials), size)
        problem.materials = problem.materials
        self.assertEqual(len(problem.materials), size)

    def test_reverse_pointers(self):
        problem = self.simple_problem
        complements = list(problem.cells[99].cells_complementing_this)
        self.assertIn(problem.cells[5], complements)
        self.assertEqual(len(complements), 1)
        cells = list(problem.materials[1].cells)
        self.assertIn(problem.cells[1], cells)
        self.assertEqual(len(cells), 1)
        cells = list(problem.surfaces[1005].cells)
        self.assertIn(problem.cells[2], problem.surfaces[1005].cells)
        self.assertEqual(len(cells), 2)

    def test_surface_card_pass_through(self):
        problem = montepy.read_input("tests/inputs/test_surfaces.imcnp")
        surf = problem.surfaces[1]
        # Test input pass through
        answer = ["1 -2 SO -5"]
        self.assertEqual(surf.format_for_mcnp_input((6, 2, 0)), answer)
        # Test changing periodic surface
        new_prob = copy.deepcopy(problem)
        surf = new_prob.surfaces[1]
        new_prob.surfaces[2].number = 5
        self.assertEqual(int(surf.format_for_mcnp_input((6, 2, 0))[0].split()[1]), -5)
        # Test changing transform
        new_prob = copy.deepcopy(problem)
        surf = new_prob.surfaces[4]
        surf.transform.number = 5
        self.assertEqual(int(surf.format_for_mcnp_input((6, 2, 0))[0].split()[1]), 5)
        # test changing surface constants
        new_prob = copy.deepcopy(problem)
        surf = new_prob.surfaces[4]
        surf.location = 2.5
        self.assertEqual(
            float(surf.format_for_mcnp_input((6, 2, 0))[0].split()[-1]), 2.5
        )

    def test_surface_broken_link(self):
        with self.assertRaises(montepy.errors.MalformedInputError):
            montepy.read_input("tests/inputs/test_broken_surf_link.imcnp")
        with self.assertRaises(montepy.errors.MalformedInputError):
            montepy.read_input("tests/inputs/test_broken_transform_link.imcnp")

    def test_material_broken_link(self):
        with self.assertRaises(montepy.errors.BrokenObjectLinkError):
            problem = montepy.read_input("tests/inputs/test_broken_mat_link.imcnp")

    def test_cell_surf_broken_link(self):
        with self.assertRaises(montepy.errors.BrokenObjectLinkError):
            problem = montepy.read_input(
                "tests/inputs/test_broken_cell_surf_link.imcnp"
            )

    def test_cell_complement_broken_link(self):
        with self.assertRaises(montepy.errors.BrokenObjectLinkError):
            problem = montepy.read_input("tests/inputs/test_broken_complement.imcnp")

    def test_cell_card_pass_through(self):
        problem = copy.deepcopy(self.simple_problem)
        cell = problem.cells[1]
        # test input pass-through
        answer = [
            "C cells",
            "c",
            "1 1 20",
            "         -1000  $ dollar comment",
            "        imp:n,p=1 U=350 trcl=5 ",
        ]
        self.assertEqual(cell.format_for_mcnp_input((6, 2, 0)), answer)
        # test surface change
        new_prob = copy.deepcopy(problem)
        new_prob.surfaces[1000].number = 5
        cell = new_prob.cells[1]
        output = cell.format_for_mcnp_input((6, 2, 0))
        print(output)
        self.assertEqual(int(output[3].split("$")[0]), -5)
        # test mass density printer
        cell.mass_density = 10.0
        with self.assertWarns(LineExpansionWarning):
            output = cell.format_for_mcnp_input((6, 2, 0))
        print(output)
        self.assertAlmostEqual(float(output[2].split()[2]), -10)
        # ensure that surface number updated
        # Test material number change
        new_prob = copy.deepcopy(problem)
        new_prob.materials[1].number = 5
        cell = new_prob.cells[1]
        output = cell.format_for_mcnp_input((6, 2, 0))
        self.assertEqual(int(output[2].split()[1]), 5)

    def test_thermal_scattering_pass_through(self):
        problem = copy.deepcopy(self.simple_problem)
        mat = problem.materials[3]
        therm = mat.thermal_scattering
        mat.number = 5
        self.assertEqual(
            therm.format_for_mcnp_input((6, 2, 0)), ["MT5 lwtr.23t h-zr.20t h/zr.28t"]
        )

    def test_cutting_comments_parse(self):
        problem = montepy.read_input("tests/inputs/breaking_comments.imcnp")
        comments = problem.cells[1].comments
        self.assertEqual(len(comments), 3)
        self.assertIn("this is a cutting comment", list(comments)[2].contents)
        comments = problem.materials[2].comments
        self.assertEqual(len(comments), 2)

    def test_cutting_comments_print_no_mutate(self):
        problem = montepy.read_input("tests/inputs/breaking_comments.imcnp")
        cell = problem.cells[1]
        output = cell.format_for_mcnp_input((6, 2, 0))
        self.assertEqual(len(output), 5)
        self.assertEqual("c this is a cutting comment", output[3])
        material = problem.materials[2]
        output = material.format_for_mcnp_input((6, 2, 0))
        self.assertEqual(len(output), 5)
        self.assertEqual("c          26057.80c        2.12", output[3])

    def test_cutting_comments_print_mutate(self):
        problem = montepy.read_input("tests/inputs/breaking_comments.imcnp")
        cell = problem.cells[1]
        cell.number = 8
        output = cell.format_for_mcnp_input((6, 2, 0))
        print(output)
        self.assertEqual(len(output), 5)
        self.assertEqual("c this is a cutting comment", output[3])
        material = problem.materials[2]
        material.number = 5
        output = material.format_for_mcnp_input((6, 2, 0))
        print(output)
        self.assertEqual(len(output), 5)
        self.assertEqual("c          26057.80c        2.12", output[3])

    def test_comments_setter(self):
        cell = copy.deepcopy(self.simple_problem.cells[1])
        comment = self.simple_problem.surfaces[1000].comments[0]
        cell.leading_comments = [comment]
        self.assertEqual(cell.comments[0], comment)
        cell.leading_comments = comment
        self.assertEqual(cell.comments[0], comment)
        with self.assertRaises(TypeError):
            cell.leading_comments = [5]
        with self.assertRaises(TypeError):
            cell.leading_comments = 5

    def test_problem_linker(self):
        cell = montepy.Cell()
        with self.assertRaises(TypeError):
            cell.link_to_problem(5)

    def test_importance_parsing(self):
        problem = self.importance_problem
        cell = problem.cells[1]
        self.assertEqual(cell.importance.neutron, 1.0)
        self.assertEqual(cell.importance.photon, 1.0)
        self.assertEqual(cell.importance.electron, 0.0)
        problem = self.simple_problem
        cell = problem.cells[1]
        self.assertEqual(cell.importance.neutron, 1.0)
        self.assertEqual(cell.importance.photon, 1.0)

    def test_importance_format_unmutated(self):
        imp = self.importance_problem.cells._importance
        output = imp.format_for_mcnp_input((6, 2, 0))
        print(output)
        self.assertEqual(len(output), 2)
        self.assertEqual("imp:n,p 1 1 1 0 3", output[0])
        self.assertEqual("imp:e   0 0 0 1 2", output[1])

    def test_importance_format_mutated(self):
        problem = copy.deepcopy(self.importance_problem)
        imp = problem.cells._importance
        problem.cells[1].importance.neutron = 0.5
        with self.assertWarns(LineExpansionWarning):
            output = imp.format_for_mcnp_input((6, 2, 0))
        print(output)
        self.assertEqual(len(output), 3)
        self.assertIn("imp:n 0.5 1 1 0 3", output)

    def test_importance_write_unmutated(self):
        out_file = "test_import_unmute"
        try:
            self.importance_problem.write_to_file(out_file)
            found_np = False
            found_e = False
            with open(out_file, "r") as fh:
                for line in fh:
                    print(line.rstrip())
                    if "imp:n,p 1" in line:
                        found_np = True
                    elif "imp:e" in line:
                        found_e = True
            self.assertTrue(found_np)
            self.assertTrue(found_e)
        finally:
            try:
                os.remove(out_file)
            except FileNotFoundError:
                pass

    def test_importance_write_mutated(self):
        out_file = "test_import_mute"
        problem = copy.deepcopy(self.importance_problem)
        problem.cells[1].importance.neutron = 0.5
        try:
            with self.assertWarns(LineExpansionWarning):
                problem.write_to_file(out_file)
            found_n = False
            found_e = False
            with open(out_file, "r") as fh:
                for line in fh:
                    print(line.rstrip())
                    if "imp:n 0.5" in line:
                        found_n = True
                    elif "imp:e" in line:
                        found_e = True
            self.assertTrue(found_n)
            self.assertTrue(found_e)
        finally:
            try:
                os.remove(out_file)
            except FileNotFoundError:
                pass

    def test_importance_write_cell(self):
        for state in ["no change", "new unmutated cell", "new mutated cell"]:
            out_file = "test_import_cell"
            problem = copy.deepcopy(self.importance_problem)
            if "new" in state:
                cell = copy.deepcopy(problem.cells[5])
                cell.number = 999
                problem.cells.append(cell)
            problem.print_in_data_block["imp"] = False
            try:
                if "new" in state:
                    with self.assertWarns(LineExpansionWarning):
                        problem.write_to_file(out_file)
                else:
                    problem.write_to_file(out_file)
                found_np = False
                found_e = False
                found_data_np = False
                with open(out_file, "r") as fh:
                    for line in fh:
                        print(line.rstrip())
                        if "imp:n,p=1" in line:
                            found_np = True
                        elif "imp:e=1" in line:
                            found_e = True
                        elif "imp:e 1" in line.lower():
                            found_data_np = True
                self.assertTrue(found_np)
                self.assertTrue(found_e)
                self.assertTrue(not found_data_np)
            finally:
                try:
                    os.remove(out_file)
                except FileNotFoundError:
                    pass

    def test_importance_write_data(self):
        out_file = "test_import_data"
        problem = copy.deepcopy(self.simple_problem)
        problem.print_in_data_block["imp"] = True
        try:
            problem.write_to_file(out_file)
            found_n = False
            found_p = False
            with open(out_file, "r") as fh:
                for line in fh:
                    print(line.rstrip())
                    if "imp:n 1" in line:
                        found_n = True
                    if "imp:p 1 0.5" in line:
                        found_p = True
            self.assertTrue(found_n)
            self.assertTrue(found_p)
        finally:
            try:
                os.remove(out_file)
            except FileNotFoundError:
                pass

    def test_avoid_blank_cell_modifier_write(self):
        out_file = "test_modifier_data"
        problem = copy.deepcopy(self.simple_problem)
        problem.print_in_data_block["U"] = True
        problem.print_in_data_block["FILL"] = True
        problem.print_in_data_block["LAT"] = True
        problem.cells[5].fill.transform = None
        problem.cells[5].fill.universe = None
        problem.cells[1].universe = problem.universes[0]
        for cell in problem.cells:
            del cell.volume
        problem.cells.allow_mcnp_volume_calc = True
        try:
            problem.write_to_file(out_file)
            found_universe = False
            found_lattice = False
            found_vol = False
            found_fill = False
            with open(out_file, "r") as fh:
                for line in fh:
                    print(line.rstrip())
                    if "U " in line:
                        found_universe = True
                    if "LAT " in line:
                        found_lat = True
                    if "FILL " in line:
                        found_fill = True
                    if "VOL " in line:
                        found_vol = True
            self.assertTrue(not found_universe)
            self.assertTrue(not found_lattice)
            self.assertTrue(not found_vol)
            self.assertTrue(not found_fill)
        finally:
            try:
                os.remove(out_file)
            except FileNotFoundError:
                pass

    def test_set_mode(self):
        problem = copy.deepcopy(self.importance_problem)
        problem.set_mode("e p")
        particles = {Particle.ELECTRON, Particle.PHOTON}
        self.assertEqual(len(problem.mode), 2)
        for part in particles:
            self.assertIn(part, problem.mode)

    def test_set_equal_importance(self):
        problem = copy.deepcopy(self.importance_problem)
        problem.cells.set_equal_importance(0.5, [5])
        for cell in problem.cells:
            for particle in problem.mode:
                if cell.number != 5:
                    print(cell.number, particle)
                    self.assertEqual(cell.importance[particle], 0.5)
        for particle in problem.mode:
            self.assertEqual(problem.cells[5].importance.neutron, 0.0)
        problem.cells.set_equal_importance(0.75, [problem.cells[99]])
        for cell in problem.cells:
            for particle in problem.mode:
                if cell.number != 99:
                    print(cell.number, particle)
                    self.assertEqual(cell.importance[particle], 0.75)
        for particle in problem.mode:
            self.assertEqual(problem.cells[99].importance.neutron, 0.0)
        with self.assertRaises(TypeError):
            problem.cells.set_equal_importance("5", [5])
        with self.assertRaises(ValueError):
            problem.cells.set_equal_importance(-0.5, [5])
        with self.assertRaises(TypeError):
            problem.cells.set_equal_importance(5, "a")
        with self.assertRaises(TypeError):
            problem.cells.set_equal_importance(5, ["a"])

    def test_check_volume_calculated(self):
        self.assertTrue(not self.simple_problem.cells[1].volume_mcnp_calc)

    def test_redundant_volume(self):
        with self.assertRaises(montepy.errors.MalformedInputError):
            montepy.read_input(
                os.path.join("tests", "inputs", "test_vol_redundant.imcnp")
            )

    def test_delete_vol(self):
        problem = copy.deepcopy(self.simple_problem)
        del problem.cells[1].volume
        self.assertTrue(not problem.cells[1].volume_is_set)

    def test_enable_mcnp_vol_calc(self):
        problem = copy.deepcopy(self.simple_problem)
        problem.cells.allow_mcnp_volume_calc = True
        self.assertTrue(problem.cells.allow_mcnp_volume_calc)
        self.assertNotIn("NO", str(problem.cells._volume))
        problem.cells.allow_mcnp_volume_calc = False
        self.assertIn("NO", str(problem.cells._volume))
        with self.assertRaises(TypeError):
            problem.cells.allow_mcnp_volume_calc = 5

    def test_cell_multi_volume(self):
        in_str = "1 0 -1 VOL=1 VOL 5"
        with self.assertRaises(ValueError):
            cell = montepy.Cell(
                Input([in_str], montepy.input_parser.block_type.BlockType.CELL)
            )

    def test_universe_cell_parsing(self):
        problem = self.simple_problem
        answers = [350] + [0] * 4
        for cell, answer in zip(problem.cells, answers):
            print(cell, answer)
            self.assertEqual(cell.universe.number, answer)

    def test_universe_fill_data_parsing(self):
        problem = montepy.read_input(
            os.path.join("tests", "inputs", "test_universe_data.imcnp")
        )
        answers = [350, 0, 0, 1]
        for cell, answer in zip(problem.cells, answers):
            print(cell, answer)
            self.assertEqual(cell.universe.number, answer)
        for cell in problem.cells:
            print(cell)
            if cell.number != 99:
                self.assertTrue(not cell.not_truncated)
            else:
                self.assertTrue(cell.not_truncated)
        self.assertTrue(problem.cells[99].not_truncated)
        answers = [None, None, 350, None, None]
        for cell, answer in zip(problem.cells, answers):
            print(cell.number, cell.fill.universe, answer)
            if answer is None:
                self.assertIsNone(cell.fill.universe)
            else:
                self.assertTrue(cell.fill.universe, answer)

    def test_universe_cells(self):
        answers = {350: [1], 0: [2, 3, 5], 1: [99]}
        problem = montepy.read_input(
            os.path.join("tests", "inputs", "test_universe_data.imcnp")
        )
        for uni_number, cell_answers in answers.items():
            for cell, answer in zip(problem.universes[uni_number].cells, cell_answers):
                self.assertEqual(cell.number, answer)

    def test_cell_not_truncate_setter(self):
        problem = copy.deepcopy(self.simple_problem)
        cell = problem.cells[1]
        cell.not_truncated = True
        self.assertTrue(cell.not_truncated)
        with self.assertRaises(ValueError):
            cell = problem.cells[2]
            cell.not_truncated = True

    def test_universe_setter(self):
        problem = copy.deepcopy(self.simple_problem)
        universe = problem.universes[350]
        cell = problem.cells[3]
        cell.universe = universe
        self.assertEqual(cell.universe, universe)
        self.assertEqual(cell.universe.number, 350)
        with self.assertRaises(TypeError):
            cell.universe = 5

    def test_universe_cell_formatter(self):
        problem = copy.deepcopy(self.simple_problem)
        universe = problem.universes[350]
        cell = problem.cells[3]
        cell.universe = universe
        cell.not_truncated = True
        with self.assertWarns(LineExpansionWarning):
            output = cell.format_for_mcnp_input((6, 2, 0))
        self.assertIn("U=-350", " ".join(output))

    def test_universe_data_formatter(self):
        problem = montepy.read_input(
            os.path.join("tests", "inputs", "test_universe_data.imcnp")
        )
        # test unmutated
        output = problem.cells._universe.format_for_mcnp_input((6, 2, 0))
        print(output)
        self.assertIn("u 350 2J -1", output)
        universe = problem.universes[350]
        # test mutated
        cell = problem.cells[3]
        cell.universe = universe
        cell.not_truncated = True
        with self.assertWarns(LineExpansionWarning):
            output = problem.cells._universe.format_for_mcnp_input((6, 2, 0))
        print(output)
        self.assertIn("u 350 J -350 -1", output)
        # test appending a new mutated cell
        new_cell = copy.deepcopy(cell)
        new_cell.number = 1000
        new_cell.universe = universe
        new_cell.not_truncated = False
        problem.cells.append(new_cell)
        with self.assertWarns(LineExpansionWarning):
            output = problem.cells._universe.format_for_mcnp_input((6, 2, 0))
        print(output)
        self.assertIn("u 350 J -350 -1 J 350 ", output)
        # test appending a new UNmutated cell
        problem = montepy.read_input(
            os.path.join("tests", "inputs", "test_universe_data.imcnp")
        )
        cell = problem.cells[3]
        new_cell = copy.deepcopy(cell)
        new_cell.number = 1000
        new_cell.universe = universe
        new_cell.not_truncated = False
        # lazily implement pulling cell in from other model
        new_cell._mutated = False
        new_cell._universe._mutated = False
        problem.cells.append(new_cell)
        with self.assertWarns(LineExpansionWarning):
            output = problem.cells._universe.format_for_mcnp_input((6, 2, 0))
        print(output)
        self.assertIn("u 350 2J -1 J 350", output)

    def test_universe_number_collision(self):
        problem = montepy.read_input(
            os.path.join("tests", "inputs", "test_universe_data.imcnp")
        )
        with self.assertRaises(montepy.errors.NumberConflictError):
            problem.universes[0].number = 350

        with self.assertRaises(ValueError):
            problem.universes[350].number = 0

    def test_universe_repr(self):
        uni = self.simple_problem.universes[0]
        output = repr(uni)
        self.assertIn("Number: 0", output)
        self.assertIn("Problem: set", output)
        self.assertIn("Cells: [2", output)

    def test_lattice_format_data(self):
        problem = copy.deepcopy(self.simple_problem)
        cells = problem.cells
        cells[1].lattice = 1
        cells[99].lattice = 2
        answer = "LAT 1 2J 2"
        output = cells._lattice.format_for_mcnp_input((6, 2, 0))
        self.assertIn(answer, output[0])

    def test_lattice_push_to_cells(self):
        problem = copy.deepcopy(self.simple_problem)
        lattices = [1, 2, Jump(), Jump()]
        card = Input(
            ["Lat " + " ".join(list(map(str, lattices)))],
            montepy.input_parser.block_type.BlockType.DATA,
        )
        lattice = montepy.data_inputs.lattice_input.LatticeInput(card)
        lattice.link_to_problem(problem)
        lattice.push_to_cells()
        for cell, answer in zip(problem.cells, lattices):
            print(cell.number, answer)
            if isinstance(answer, int):
                self.assertEqual(cell.lattice.value, answer)
            else:
                self.assertIsNone(cell.lattice)

    def test_universe_problem_parsing(self):
        problem = self.universe_problem
        for cell in problem.cells:
            if cell.number == 1:
                self.assertEqual(cell.universe.number, 1)
            else:
                self.assertEqual(cell.universe.number, 0)

    def test_importance_end_repeat(self):
        problem = copy.deepcopy(self.universe_problem)
        for cell in problem.cells:
            if cell.number in {99, 5}:
                cell.importance.photon = 1.0
            else:
                cell.importance.photon = 0.0
        problem.print_in_data_block["IMP"] = True
        output = problem.cells._importance.format_for_mcnp_input((6, 2, 0))
        # OG value was 0.5 so 0.0 is correct.
        self.assertIn("imp:p 0 0.0", output)

    def test_fill_parsing(self):
        problem = self.universe_problem
        answers = [None, np.array([[[1], [0]], [[0], [1]]]), None, 1, 1]
        for cell, answer in zip(problem.cells, answers):
            if answer is None:
                self.assertIsNone(cell.fill.universe)
            elif isinstance(answer, np.ndarray):
                self.assertTrue(cell.fill.multiple_universes)
                self.assertTrue(
                    (cell.fill.min_index == np.array([0.0, 0.0, 0.0])).all()
                )
                self.assertTrue(
                    (cell.fill.max_index == np.array([1.0, 1.0, 0.0])).all()
                )
                self.assertEqual(cell.fill.universes[0][0][0].number, answer[0][0][0])
                self.assertEqual(cell.fill.universes[1][1][0].number, answer[1][1][0])
                self.assertEqual(cell.fill.transform, problem.transforms[5])
            else:
                self.assertEqual(cell.fill.universe.number, answer)

    def test_fill_transform_setter(self):
        problem = copy.deepcopy(self.universe_problem)
        transform = problem.transforms[5]
        cell = problem.cells[5]
        cell.fill.transform = transform
        self.assertEqual(cell.fill.transform, transform)
        self.assertTrue(not cell.fill.hidden_transform)
        cell.fill.transform = None
        self.assertIsNone(cell.fill.transform)
        with self.assertRaises(TypeError):
            cell.fill.transform = "hi"
        cell.fill.transform = transform
        del cell.fill.transform
        self.assertIsNone(cell.fill.transform)

    def test_fill_cell_format(self):
        problem = copy.deepcopy(self.universe_problem)
        fill = problem.cells[5].fill
        output = fill.format_for_mcnp_input((6, 2, 0))
        answer = "fill=1 (1 0.0 0.0)"
        self.assertEqual(output[0], answer)
        # test *fill
        fill.transform.is_in_degrees = True
        output = fill.format_for_mcnp_input((6, 2, 0))
        answer = "*fill=1 (1 0.0 0.0)"
        self.assertEqual(output[0], answer)
        # test changing the transform
        fill.transform.displacement_vector[0] = 2.0
        output = fill.format_for_mcnp_input((6, 2, 0))
        self.assertEqual(output[0], "*fill=1 (2 0.0 0.0)")
        # test without transform
        fill.transform = None
        answer = "fill=1 "
        output = fill.format_for_mcnp_input((6, 2, 0))
        self.assertEqual(output[0], answer)
        # test with no fill
        fill.universe = None
        output = fill.format_for_mcnp_input((6, 2, 0))
        self.assertEqual(len(output), 0)
        # test with complex universe lattice fill
        fill = problem.cells[2].fill
        output = fill.format_for_mcnp_input((6, 2, 0))
        answers = ["fill= 0:1 0:1 0:0 1 0 0 1 (5)"]
        self.assertEqual(output, answers)
        problem.print_in_data_block["FILL"] = True
        # test that complex fill is not printed in data block
        with self.assertRaises(ValueError):
            output = problem.cells._fill.format_for_mcnp_input((6, 2, 0))
        problem = copy.deepcopy(self.simple_problem)
        problem.cells[5].fill.transform = None
        problem.print_in_data_block["FILL"] = True
        output = problem.cells._fill.format_for_mcnp_input((6, 2, 0))
        self.assertEqual(output, ["FILL 4J 350 "])

    def test_universe_cells_claim(self):
        problem = copy.deepcopy(self.universe_problem)
        universe = problem.universes[1]
        universe.claim(problem.cells[2])
        self.assertEqual(problem.cells[2].universe, universe)
        universe = montepy.Universe(5)
        problem.universes.append(universe)
        universe.claim(problem.cells)
        for cell in problem.cells:
            self.assertEqual(cell.universe, universe)
        with self.assertRaises(TypeError):
            universe.claim("hi")
        with self.assertRaises(TypeError):
            universe.claim(["hi"])

    def test_universe_cells(self):
        problem = self.universe_problem
        answers = [1]
        universe = problem.universes[1]
        self.assertEqual(len(answers), len(list(universe.cells)))
        for cell, answer in zip(universe.cells, answers):
            self.assertEqual(cell.number, answer)

    def test_data_print_control_str(self):
        self.assertEqual(
            str(self.simple_problem.print_in_data_block),
            "Print data in data block: {'imp': False, 'u': False, 'fill': False, 'vol': True}",
        )

    def test_cell_validator(self):
        problem = copy.deepcopy(self.simple_problem)
        cell = problem.cells[1]
        del cell.mass_density
        with self.assertRaises(montepy.errors.IllegalState):
            cell.validate()
        cell = montepy.Cell()
        # test no geometry at all
        with self.assertRaises(montepy.errors.IllegalState):
            cell.validate()
        surf = problem.surfaces[1000]
        cell.surfaces.append(surf)
        # test surface added but geomtry not defined
        with self.assertRaises(montepy.errors.IllegalState):
            cell.validate()

    def test_importance_rewrite(self):
        out_file = "test_import_data"
        problem = copy.deepcopy(self.simple_problem)
        problem.print_in_data_block["imp"] = True
        try:
            problem.write_to_file(out_file)
            problem = montepy.read_input(out_file)
            problem.print_in_data_block["imp"] = False
            problem.write_to_file(out_file)
            found_n = False
            found_p = False
            found_vol = False
            with open(out_file, "r") as fh:
                for line in fh:
                    print(line.rstrip())
                    if "IMP:N 1 2R" in line:
                        found_n = True
                    if "IMP:P 1 0.5" in line:
                        found_p = True
                    if "vol NO 2J 1 1.5 J" in line:
                        found_vol = True
            self.assertTrue(not found_n)
            self.assertTrue(not found_p)
            self.assertTrue(found_vol)
        finally:
            try:
                os.remove(out_file)
            except FileNotFoundError:
                pass

    def test_parsing_error(self):
        in_file = os.path.join("tests", "inputs", "test_bad_syntax.imcnp")
        with self.assertRaises(montepy.errors.ParsingError):
            problem = montepy.read_input(in_file)

    def test_leading_comments(self):
        cell = copy.deepcopy(self.simple_problem.cells[1])
        leading_comments = cell.leading_comments
        self.assertIn("cells", leading_comments[0].contents)
        del cell.leading_comments
        self.assertFalse(cell.leading_comments)
        cell.leading_comments = leading_comments[0:1]
        self.assertIn("cells", cell.leading_comments[0].contents)
        self.assertEqual(len(cell.leading_comments), 1)

    def test_wrap_warning(self):
        cell = copy.deepcopy(self.simple_problem.cells[1])
        with self.assertWarns(montepy.errors.LineExpansionWarning):
            output = cell.wrap_string_for_mcnp("h" * 130, (6, 2, 0), True)
            self.assertEqual(len(output), 2)
        output = cell.wrap_string_for_mcnp("h" * 127, (6, 2, 0), True)
        self.assertEqual(len(output), 1)

    def test_expansion_warning_crash(self):
        problem = copy.deepcopy(self.simple_problem)
        cell = problem.cells[99]
        cell.material = problem.materials[1]
        cell.mass_density = 10.0
        problem.materials[1].number = 987654321
        problem.surfaces[1010].number = 123456789
        out = "bad_warning.imcnp"
        try:
            with self.assertWarns(montepy.errors.LineExpansionWarning):
                problem.write_to_file(out)
        finally:
            try:
                os.remove(out)
            except FileNotFoundError:
                pass



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import glob
import montepy
from montepy import __main__ as main
from montepy.errors import *
from unittest import TestCase
from tests import constants
import os


class TestMainRunner(TestCase):
    def test_argument_parsing(self):
        arguments = ["-c", "foo.imcnp", "bar.imcnp"]
        args = main.define_args(arguments)
        self.assertEqual(len(args.check), 2)
        self.assertIn("foo.imcnp", args.check)
        self.assertIn("bar.imcnp", args.check)

    def test_check_no_error(self):
        main.check_inputs([os.path.join("tests", "inputs", "test.imcnp")])
        with self.assertRaises(FileNotFoundError):
            main.check_inputs(["foo"])

    def test_check_warning(self):
        for bad_input in constants.BAD_INPUTS:
            with self.assertWarns(Warning):
                print(f"Testing that a warning is issued for: {bad_input}")
                main.check_inputs([os.path.join("tests", "inputs", bad_input)])
        for file in glob.glob(os.path.join("tests", "inputs", "*.imcnp")):
            if (
                os.path.basename(file) not in constants.BAD_INPUTS
                and os.path.basename(file) not in constants.IGNORE_FILES
            ):
                print(f"Testing no errors are raised for file: {file}")
                montepy.read_input(file)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import itertools as it
from montepy.utilities import *


class MCNP_InputFile:
    """
    A class to represent a distinct input file.

    .. Note::
        this is a bare bones implementation to be fleshed out in the future.

    :param path: the path to the input file
    :type path: str
    :param parent_file: the parent file for this file if any. This occurs when a "read" input is used.
    :type parent_file: str
    """

    def __init__(self, path, parent_file=None):
        self._path = path
        self._parent_file = parent_file
        self._lineno = 1
        self._fh = None

    @make_prop_pointer("_path")
    def path(self):
        """
        The path for the file.

        :rtype: str
        """
        pass

    @property
    def name(self):
        return self.path

    @make_prop_pointer("_parent_file")
    def parent_file(self):
        """
        The parent file for this file.

        This is only used when this file is pointed to by a "read" input.

        :rtype: str
        """
        pass

    @make_prop_pointer("_lineno")
    def lineno(self):
        """
        The current line number being read in the file.

        This is 1-indexed.

        :rtype: int
        """
        pass

    def open(self, mode):
        """
        Opens the underlying file, and returns self.

        This should only ever be completed from within a ``with`` statement.
        For this reason, a ``close`` functional is intentionally
        not provided.

        :param mode: the mode to open the file in
        :type mode: str
        :returns: self
        """
        self._fh = open(self.path, mode, encoding="ascii")
        return self

    def __enter__(self):
        self._fh.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        status = self._fh.__exit__(exc_type, exc_val, exc_tb)
        self._fh = None
        return status

    def __iter__(self):
        for lineno, line in enumerate(self._fh):
            self._lineno = lineno + 1
            yield line

    def read(self, size=-1):
        """ """
        if self._fh:
            ret = self._fh.read(size)
            self._lineno += ret.count("\n")
            return ret

    def readline(self, size=-1):
        """ """
        if self._fh:
            ret = self._fh.readline(size)
            self._lineno += ret.count("\n")
            return ret

    def write(self, to_write):
        if self._fh:
            self._lineno += to_write.count("\n")
            return self._fh.write(to_write)

    def __str__(self):
        return self.name



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from unittest import TestCase

import copy
from montepy.constants import DEFAULT_VERSION
from montepy.input_parser import syntax_node
import montepy
from montepy.cell import Cell
from montepy.constants import DEFAULT_VERSION
from montepy.errors import *
from montepy.input_parser.block_type import BlockType
from montepy.input_parser.mcnp_input import Input, Jump
from montepy.universe import Universe
from montepy.data_inputs.fill import Fill
from montepy.data_inputs.lattice import Lattice
from montepy.data_inputs.lattice_input import LatticeInput
from montepy.data_inputs.universe_input import UniverseInput
import numpy as np


class TestUniverseInput(TestCase):
    def setUp(self):
        list_node = syntax_node.ListNode("numbers")
        list_node.append(syntax_node.ValueNode("5", float))
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = "u"
        tree = syntax_node.SyntaxNode(
            "lattice",
            {
                "classifier": classifier,
                "seperator": syntax_node.ValueNode("=", str),
                "data": list_node,
            },
        )
        self.tree = tree
        self.universe = UniverseInput(in_cell_block=True, key="u", value=tree)

    def test_universe_card_init(self):
        card = self.universe
        self.assertEqual(card.old_number, 5)
        self.assertTrue(not card.not_truncated)
        # test bad float
        with self.assertRaises(ValueError):
            tree = copy.deepcopy(self.tree)
            tree["data"].nodes.pop()
            tree["data"].append(syntax_node.ValueNode("5.5", float))
            card = UniverseInput(in_cell_block=True, key="U", value=tree)

        # test string
        with self.assertRaises(ValueError):
            tree["data"].nodes.pop()
            tree["data"].append(syntax_node.ValueNode("hi", str))
            card = UniverseInput(in_cell_block=True, key="U", value=tree)

        # test negative universe
        tree["data"].nodes.pop()
        tree["data"].append(syntax_node.ValueNode("-3", float))
        card = UniverseInput(in_cell_block=True, key="U", value=tree)
        self.assertEqual(card.old_number, 3)
        self.assertTrue(card.not_truncated)

        universes = [1, 2, 3]
        card = Input(["U " + " ".join(list(map(str, universes)))], BlockType.DATA)
        uni_card = UniverseInput(card)
        self.assertEqual(uni_card.old_numbers, universes)

        # test jump
        card = Input(["U J"], BlockType.DATA)
        uni_card = UniverseInput(card)
        self.assertEqual(uni_card.old_numbers[0], None)

        # test bad float
        with self.assertRaises(MalformedInputError):
            card = Input(["U 5.5"], BlockType.DATA)
            uni_card = UniverseInput(card)

        # test bad str
        with self.assertRaises(MalformedInputError):
            card = Input(["U hi"], BlockType.DATA)
            uni_card = UniverseInput(card)

        # test bad negative
        card = Input(["U -2"], BlockType.DATA)
        uni_card = UniverseInput(card)

    def test_str(self):
        card = copy.deepcopy(self.universe)
        uni = Universe(5)
        card.universe = uni
        output = str(card)
        self.assertIn("u=5", output)
        output = repr(card)
        self.assertIn("UNIVERSE", output)
        self.assertIn("set_in_block: True", output)
        self.assertIn("Universe : Universe(5)", output)

    def test_merge(self):
        card = copy.deepcopy(self.universe)
        with self.assertRaises(MalformedInputError):
            card.merge(card)

    def test_universe_setter(self):
        card = copy.deepcopy(self.universe)
        uni = Universe(5)
        card.universe = uni
        self.assertEqual(card.universe, uni)
        with self.assertRaises(TypeError):
            card.universe = 5

    def test_universe_truncate_setter(self):
        card = copy.deepcopy(self.universe)
        self.assertTrue(not card.not_truncated)
        card.not_truncated = True
        self.assertTrue(card.not_truncated)
        with self.assertRaises(TypeError):
            card.not_truncated = 5


class TestUniverse(TestCase):
    def test_init(self):
        universe = Universe(5)
        self.assertEqual(universe.number, 5)
        self.assertEqual(universe.old_number, 5)
        with self.assertRaises(TypeError):
            Universe("hi")
        with self.assertRaises(ValueError):
            Universe(-1)

    def test_number_setter(self):
        universe = Universe(5)
        universe.number = 10
        self.assertEqual(universe.number, 10)
        with self.assertRaises(TypeError):
            universe.number = "hi"
        with self.assertRaises(ValueError):
            universe.number = -1


class TestLattice(TestCase):
    def setUp(self):
        list_node = syntax_node.ListNode("numbers")
        list_node.append(syntax_node.ValueNode("1", float))
        classifier = syntax_node.ClassifierNode()
        classifier.prefix = "lat"
        tree = syntax_node.SyntaxNode(
            "lattice",
            {
                "classifier": classifier,
                "seperator": syntax_node.ValueNode("=", str),
                "data": list_node,
            },
        )
        self.tree = tree
        self.lattice = LatticeInput(in_cell_block=True, key="lat", value=tree)

    def test_lattice_init(self):
        lattice = self.lattice
        self.assertEqual(lattice.lattice, Lattice(1))
        tree = copy.deepcopy(self.tree)
        with self.assertRaises(ValueError):
            tree["data"].nodes.pop()
            tree["data"].append(syntax_node.ValueNode("hi", str))
            lattice = LatticeInput(in_cell_block=True, key="lat", value=tree)
        with self.assertRaises(ValueError):
            tree["data"].nodes.pop()
            tree["data"].append(syntax_node.ValueNode("5", float))
            lattice = LatticeInput(in_cell_block=True, key="lat", value=tree)
        lattices = [1, 2, None, None]
        input = Input(["Lat " + " ".join(list(map(str, lattices)))], BlockType.DATA)
        lattice = LatticeInput(input)
        for answer, lattice in zip(lattices, lattice._lattice):
            self.assertEqual(Lattice(answer), lattice.value)
        with self.assertRaises(MalformedInputError):
            card = Input(["Lat 3"], BlockType.DATA)
            LatticeInput(card)
        with self.assertRaises(MalformedInputError):
            card = Input(["Lat str"], BlockType.DATA)
            LatticeInput(card)

    def test_lattice_setter(self):
        lattice = copy.deepcopy(self.lattice)
        lattice.lattice = Lattice(2)
        self.assertEqual(Lattice(2), lattice.lattice)
        lattice.lattice = 1
        self.assertEqual(Lattice(1), lattice.lattice)
        lattice.lattice = None
        self.assertIsNone(lattice.lattice)
        with self.assertRaises(TypeError):
            lattice.lattice = "hi"

        with self.assertRaises(ValueError):
            lattice.lattice = -1

    def test_lattice_deleter(self):
        lattice = self.lattice
        del lattice.lattice
        self.assertIsNone(lattice.lattice)

    def test_lattice_merge(self):
        lattice = self.lattice
        with self.assertRaises(MalformedInputError):
            lattice.merge(lattice)

    def test_lattice_cell_format(self):
        lattice = self.lattice
        output = lattice.format_for_mcnp_input(DEFAULT_VERSION)
        self.assertIn("lat=1", output[0])
        lattice.lattice = None
        output = lattice.format_for_mcnp_input(DEFAULT_VERSION)
        self.assertEqual(output, [])

    def test_lattice_repr(self):
        lattice = self.lattice
        out = repr(lattice)
        self.assertIn("in_cell: True", out)
        self.assertIn("set_in_block: True", out)
        self.assertIn("Lattice_values : Lattice.HEXAHEDRA", out)


class TestFill(TestCase):
    def setUp(self):
        list_node = syntax_node.ListNode("num")
        list_node.append(syntax_node.ValueNode("5", float))
        tree = syntax_node.SyntaxNode(
            "fill",
            {
                "classifier": "",
                "seperator": syntax_node.ValueNode("=", str),
                "data": list_node,
            },
        )
        self.tree = tree
        self.simple_fill = Fill(in_cell_block=True, key="fill", value=tree)

    def test_complex_transform_fill_init(self):
        input = Input(["1 0 -1 *fill=1 (1.5 0.0 0.0)"], BlockType.CELL)
        cell = Cell(input)
        fill = cell.fill
        self.assertTrue(fill.hidden_transform)
        self.assertIsNone(fill.universes)
        self.assertEqual(fill.old_universe_number, 1)
        self.assertEqual(len(fill.transform.displacement_vector), 3)
        self.assertTrue(fill.transform.is_in_degrees)
        input = Input(["1 0 -1 fill=1 (1.5 0.0 0.0)"], BlockType.CELL)
        cell = Cell(input)
        fill = cell.fill
        self.assertTrue(not fill.transform.is_in_degrees)
        input = Input(["1 0 -1 fill=5 (3)"], BlockType.CELL)
        cell = Cell(input)
        fill = cell.fill
        self.assertTrue(not fill.hidden_transform)
        self.assertEqual(fill.old_universe_number, 5)
        self.assertEqual(fill.old_transform_number, 3)
        # test bad string
        with self.assertRaises(ValueError):
            input = Input(["1 0 -1 fill=hi"], BlockType.CELL)
            cell = Cell(input)
            fill = cell.fill
        with self.assertRaises(ValueError):
            input = Input(["1 0 -1 fill=1 (hi)"], BlockType.CELL)
            cell = Cell(input)
            fill = cell.fill
        # test negative universe
        with self.assertRaises(ValueError):
            input = Input(["1 0 -1 fill=-5"], BlockType.CELL)
            cell = Cell(input)
            fill = cell.fill
        with self.assertRaises(ValueError):
            input = Input(["1 0 -1 fill=5 (-5)"], BlockType.CELL)
            cell = Cell(input)
            fill = cell.fill
        with self.assertRaises(ValueError):
            input = Input(["1 0 -1 fill=5 1 0 0"], BlockType.CELL)
            cell = Cell(input)
            fill = cell.fill

    def test_complicated_lattice_fill_init(self):
        input = Input(["1 0 -1 fill=0:1 0:1 0:1 1 2 3 4 5 6 7 8"], BlockType.CELL)
        cell = Cell(input)
        fill = cell.fill
        self.assertIsNone(fill.universe)
        self.assertEqual(fill.min_index[0], 0)
        self.assertEqual(fill.max_index[2], 1)
        answer = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        self.assertTrue((fill.old_universe_numbers == answer).all())
        # test string universe
        with self.assertRaises(ValueError):
            input = Input(["1 0 -1 fill=0:1 0:1 0:1 hi"], BlockType.CELL)
            cell = Cell(input)
            fill = cell.fill
        # test string index
        with self.assertRaises(ValueError):
            input = Input(["1 0 -1 fill=0:1 hi:1 0:1 hi"], BlockType.CELL)
            cell = Cell(input)
            fill = cell.fill
        # test negative universe
        with self.assertRaises(ValueError):
            input = Input(["1 0 -1 fill=0:1 0:1 0:1 -1"], BlockType.CELL)
            cell = Cell(input)
            fill = cell.fill
        # test inverted bounds
        with self.assertRaises(ValueError):
            input = Input(["1 0 -1 fill=0:1 1:0 0:1 1 2 3 4 5 6 7 8"], BlockType.CELL)
            cell = Cell(input)
            fill = cell.fill
        # test float bounds
        with self.assertRaises(ValueError):
            input = Input(["1 0 -1 fill=0:1 0:1.5 0:1 1 2 3 4 5 6 7 8"], BlockType.CELL)
            cell = Cell(input)
            fill = cell.fill

    def test_data_fill_init(self):
        card = Input(["FiLl 1 2 3 4"], BlockType.DATA)
        fill = Fill(card)
        answer = [1, 2, 3, 4]
        self.assertEqual(fill.old_universe_numbers, answer)
        # jump
        card = Input(["FiLl 1 2J 4"], BlockType.DATA)
        fill = Fill(card)
        answer = [1, None, None, 4]
        self.assertEqual(fill.old_universe_numbers, answer)
        # test negative universe
        with self.assertRaises(MalformedInputError):
            card = Input(["FiLl 1 -2 3 4"], BlockType.DATA)
            fill = Fill(card)
        # test string universe
        with self.assertRaises(MalformedInputError):
            card = Input(["FiLl 1 foo"], BlockType.DATA)
            fill = Fill(card)

    def test_fill_universe_setter(self):
        list_node = syntax_node.ListNode("num")
        list_node.append(syntax_node.ValueNode("5", float))
        value = syntax_node.SyntaxNode(
            "fill",
            {
                "classifier": "",
                "seperator": syntax_node.ValueNode("=", str),
                "data": list_node,
            },
        )
        fill = copy.deepcopy(self.simple_fill)
        uni = montepy.Universe(6)
        fill.universe = uni
        self.assertEqual(fill.universe.number, uni.number)
        self.assertIsNone(fill.universes)
        fill.universe = None
        self.assertIsNone(fill.universe)
        fill.universe = uni
        del fill.universe
        self.assertIsNone(fill.universe)
        with self.assertRaises(TypeError):
            fill.universe = "hi"
        with self.assertRaises(TypeError):
            fill.multiple_universes = "hi"
        fill.multiple_universes = True
        with self.assertRaises(ValueError):
            fill.universe = uni

    def test_fill_universes_setter(self):
        input = Input(["1 0 -1 fill=0:1 0:1 0:1 1 2 3 4 5 6 7 8"], BlockType.CELL)
        cell = Cell(input)
        fill = cell.fill
        uni = montepy.Universe(10)
        fill_array = np.array([[[uni, uni], [uni, uni]], [[uni, uni], [uni, uni]]])
        fill.universes = fill_array
        self.assertTrue((fill.universes == fill_array).all())
        del fill.universes
        self.assertIsNone(fill.universes)
        with self.assertRaises(TypeError):
            fill.universes = "hi"
        fill.multiple_universes = False
        with self.assertRaises(ValueError):
            fill.universes = fill_array

    def test_fill_str(self):
        input = Input(["1 0 -1 fill=0:1 0:1 0:1 1 2 3 4 5 6 7 8"], BlockType.CELL)
        cell = Cell(input)
        fill = cell.fill
        output = str(fill)
        self.assertIn("Fill", output)
        output = repr(fill)
        self.assertIn("Fill", output)

    def test_fill_merge(self):
        card = Input(["FiLl 1 2 3 4"], BlockType.DATA)
        fill1 = Fill(card)
        fill2 = Fill(card)
        with self.assertRaises(MalformedInputError):
            fill1.merge(fill2)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.data_inputs.data_input import DataInputAbstract
from montepy.input_parser.thermal_parser import ThermalParser
from montepy import mcnp_object
from montepy.errors import *
from montepy.utilities import *
import montepy


class ThermalScatteringLaw(DataInputAbstract):
    """
    Class to hold MT Inputs

    This is designed to be called two ways.
    The first is with a read input file using input_card, comment
    The second is after a read with a material and a comment (using named inputs)

    :param input: the Input object representing this data input
    :type input: Input
    :param material: the parent Material object that owns this
    :type material: Material
    """

    _parser = ThermalParser()

    def __init__(self, input="", material=None):
        self._old_number = self._generate_default_node(int, -1)
        self._parent_material = None
        self._scattering_laws = []
        super().__init__(input)
        if input:
            self._old_number = self._input_number
            self._scattering_laws = self._tree["data"].nodes
        else:
            if material:
                self._parent_material = material

    @staticmethod
    def _class_prefix():
        return "mt"

    @staticmethod
    def _has_number():
        return True

    @staticmethod
    def _has_classifier():
        return 0

    @make_prop_val_node("_old_number")
    def old_number(self):
        """
        The material number from the file

        :rtype: int
        """
        pass

    @property
    def parent_material(self):
        """
        The Material object this is tied to.

        :rtype: Material
        """
        return self._parent_material

    @property
    def thermal_scattering_laws(self):
        """
        The thermal scattering laws to use for this material as strings.

        :rtype: list
        """
        ret = []
        for law in self._scattering_laws:
            ret.append(law.value)
        return ret

    @thermal_scattering_laws.setter
    def thermal_scattering_laws(self, laws):
        if not isinstance(laws, list):
            raise TypeError("thermal_scattering_laws must be a list")
        for law in laws:
            if not isinstance(law, str):
                raise TypeError(
                    f"element {law} in thermal_scattering_laws must be a string"
                )
        self._scattering_laws.clear()
        for law in laws:
            self._scattering_laws.append(self._generate_default_node(str, law))

    def add_scattering_law(self, law):
        """
        Adds the requested scattering law to this material

        :param law: the thermal scattering law to add.
        :type law: str
        """
        self._scattering_laws.append(self._generate_default_node(str, law))

    def validate(self):
        if len(self._scattering_laws) == 0:
            if self.parent_material:
                message = f"No thermal scattering laws given for MT{self.parent_material.number}."
            else:
                message = f"No thermal scattering laws given for thermal scattering {hex(id(self))}"
            raise IllegalState(message)

    def _update_values(self):
        if self.parent_material:
            self._tree["classifier"].number.value = self.parent_material.number
        else:
            raise MalformedInputError(
                self._input,
                f"MT{self.old_number} input is detached from a parent material",
            )

    def update_pointers(self, data_inputs):
        """
        Updates pointer to the thermal scattering data

        :param data_inputs: a list of the data inputs in the problem
        :type data_inputs: list
        """
        found = False
        for input in data_inputs:
            if isinstance(input, montepy.data_inputs.material.Material):
                if input.number == self.old_number:
                    found = True
                    self._parent_material = input

        if not found:
            raise MalformedInputError(
                self._input, "MT input is detached from a parent material"
            )

    def __str__(self):
        return f"THERMAL SCATTER: {self.thermal_scattering_laws}"

    def __repr__(self):
        return f"THERMAL SCATTER: material: {self.parent_material}, old_num: {self.old_number}, scatter: {self.thermal_scattering_laws}"



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from enum import unique, Enum


# @unique
class SurfaceType(str, Enum):
    """
    An enumeration of the surface types allowed.

    :param value: The shorthand used by MCNP
    :type value: str
    :param description: The human readable description of the surface.
    :type description: str
    """

    def __new__(cls, value, description):
        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.description = description
        return obj

    def __str__(self):
        return self.value

    # planes
    P = ("P", "general plane")
    PX = ("PX", "plane normal to x-axis")
    PY = ("PY", "plane normal to y-axis")
    PZ = ("PZ", "plane normal to z-axis")
    # spheres
    SO = ("SO", "sphere centered at origin")
    S = ("S", "general sphere")
    SX = ("SX", "sphere centered on x-axis")
    SY = ("SY", "sphere centered on y-axis")
    SZ = ("SZ", "sphere centered on z-axis")
    # cylinder
    C_X = ("C/X", "cylinder parallel to x-axis")
    C_Y = ("C/Y", "cylinder parallel to y-axis")
    C_Z = ("C/Z", "cylinder parallel to z-axis")
    CX = ("CX", "cylinder on x-axis")
    CY = ("CY", "cylinder on y-axis")
    CZ = ("CZ", "cylinder on z-axis")
    # cone
    K_X = ("K/X", "cone parallel to x-axis")
    K_Y = ("K/Y", "cone parallel to y-axis")
    K_Z = ("K/Z", "cone parallel to z-axis")
    KX = ("KX", "cone on x-axis")
    KY = ("KY", "cone on y-axis")
    KZ = ("KZ", "cone on z-axis")
    # generalized 3D conics
    SQ = ("SQ", "ellipsoid, hyperboloid, or paraboloid parallel to an axis")
    GQ = (
        "GQ",
        "cylinder, cone, ellipsoid hyperboloid, or parabaloid not parallel to an axis",
    )
    # Torus
    TX = ("TX", "elliptical torus parallel to x-axis")
    TY = ("TY", "elliptical torus parallel to y-axis")
    TZ = ("TZ", "elliptical torus parallel to z-axis")
    # by points
    X = ("X", "axisymmetric surface defined by points")
    Y = ("Y", "axisymmetric surface defined by points")
    Z = ("Z", "axisymmetric surface defined by points")
    # macrobodies
    BOX = ("BOX", "orthogonal box")
    RPP = ("RPP", "rectangular parallelepiped")
    SPH = ("SPH", "sphere")
    RCC = ("RCC", "right circular cylinder")
    RHP = ("RHP", "right hexagonal prism")
    HEX = ("HEX", "right hexagonal prism")
    REC = ("REC", "right elliptical cylinder")
    TRC = ("TRC", "truncated right-angle cone")
    ELL = ("ELL", "ellipsoid")
    WED = ("WED", "wedge")
    ARB = ("ARB", "arbitrary polyhedron")



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.data_inputs.isotope import Isotope
from montepy.input_parser.syntax_node import PaddingNode, ValueNode
from montepy.utilities import make_prop_val_node


def _enforce_positive(self, val):
    if val <= 0:
        raise ValueError(f"material component fraction must be > 0. {val} given.")


class MaterialComponent:
    """
    A class to represent a single component in a material.

    For example: this may be H-1 in water: like 1001.80c — 0.6667

    :param isotope: the Isotope object representing this isotope
    :type isotope: Isotope
    :param fraction: the fraction of this component in the material
    :type fraction: ValueNode
    """

    def __init__(self, isotope, fraction):
        if not isinstance(isotope, Isotope):
            raise TypeError(f"Isotope must be an Isotope. {isotope} given")
        if isinstance(fraction, (float, int)):
            fraction = ValueNode(str(fraction), float, padding=PaddingNode(" "))
        elif not isinstance(fraction, ValueNode) or not isinstance(
            fraction.value, float
        ):
            raise TypeError(f"fraction must be float ValueNode. {fraction} given.")
        self._isotope = isotope
        self._tree = fraction
        if fraction.value < 0:
            raise ValueError(f"Fraction must be > 0. {fraction.value} given.")
        self._fraction = fraction

    @property
    def isotope(self):
        """
        The isotope for this material_component

        :rtype: Isotope
        """
        return self._isotope

    @make_prop_val_node("_fraction", (float, int), float, _enforce_positive)
    def fraction(self):
        """
        The fraction of the isotope for this component

        :rtype: float
        """
        pass

    def __str__(self):
        return f"{self.isotope} {self.fraction}"

    def __repr__(self):
        return f"{self.isotope} {self.fraction}"



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
""" MontePy is a library for reading, editing, and writing MCNP input files.

This creates a semantic understanding of the MCNP input file.
start by running montepy.read_input().

You will receive an MCNP_Problem object that you will interact with.
"""

from . import input_parser
from . import constants
import importlib.metadata
from .input_parser.input_reader import read_input
from montepy.cell import Cell
from montepy.mcnp_problem import MCNP_Problem
from montepy.data_inputs.material import Material
from montepy.data_inputs.transform import Transform
from montepy.geometry_operators import Operator
from montepy import geometry_operators
from montepy.input_parser.mcnp_input import Jump
from montepy.particle import Particle
from montepy.surfaces.surface_type import SurfaceType
from montepy.universe import Universe
import sys

__version__ = "0.2.5"

# enable deprecated warnings for users
if not sys.warnoptions:
    import os, warnings

    warnings.simplefilter("default")  # Change the filter in this process
    os.environ["PYTHONWARNINGS"] = "default"  # Also affect subprocesses
